#!/usr/bin/env python3
# Copyright (c) 2016-2022 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Test compact blocks (BIP 152)."""
import random

from test_framework.blocktools import (
    COINBASE_MATURITY,
    NORMAL_GBT_REQUEST_PARAMS,
    add_witness_commitment,
    create_block,
)
from test_framework.messages import (
    BlockTransactions,
    BlockTransactionsRequest,
    CBlock,
    CBlockHeader,
    CInv,
    COutPoint,
    CTransaction,
    CTxIn,
    CTxInWitness,
    CTxOut,
    from_hex,
    HeaderAndShortIDs,
    MSG_BLOCK,
    MSG_CMPCT_BLOCK,
    MSG_WITNESS_FLAG,
    P2PHeaderAndShortIDs,
    PrefilledTransaction,
    calculate_shortid,
    msg_block,
    msg_blocktxn,
    msg_cmpctblock,
    msg_getblocktxn,
    msg_getdata,
    msg_getheaders,
    msg_headers,
    msg_inv,
    msg_no_witness_block,
    msg_no_witness_blocktxn,
    msg_sendcmpct,
    msg_sendheaders,
    msg_tx,
    ser_uint256,
    tx_from_hex,
)
from test_framework.p2p import (
    P2PInterface,
    p2p_lock,
)
from test_framework.script import (
    CScript,
    OP_DROP,
    OP_TRUE,
)
from test_framework.test_framework import BitcoinTestFramework
from test_framework.util import (
    assert_equal,
    softfork_active,
)
from test_framework.wallet import MiniWallet


# TestP2PConn: A peer we use to send messages to bitcoind, and store responses.
class TestP2PConn(P2PInterface):
    def __init__(self):
        super().__init__()
        self.last_sendcmpct = []
        self.block_announced = False
        # Store the hashes of blocks we've seen announced.
        # This is for synchronizing the p2p message traffic,
        # so we can eg wait until a particular block is announced.
        self.announced_blockhashes = set()

    def on_sendcmpct(self, message):
        self.last_sendcmpct.append(message)

    def on_cmpctblock(self, message):
        self.block_announced = True
        self.last_message["cmpctblock"].header_and_shortids.header.calc_sha256()
        self.announced_blockhashes.add(self.last_message["cmpctblock"].header_and_shortids.header.sha256)

    def on_headers(self, message):
        self.block_announced = True
        for x in self.last_message["headers"].headers:
            x.calc_sha256()
            self.announced_blockhashes.add(x.sha256)

    def on_inv(self, message):
        for x in self.last_message["inv"].inv:
            if x.type == MSG_BLOCK:
                self.block_announced = True
                self.announced_blockhashes.add(x.hash)

    # Requires caller to hold p2p_lock
    def received_block_announcement(self):
        return self.block_announced

    def clear_block_announcement(self):
        with p2p_lock:
            self.block_announced = False
            self.last_message.pop("inv", None)
            self.last_message.pop("headers", None)
            self.last_message.pop("cmpctblock", None)

    def clear_getblocktxn(self):
        with p2p_lock:
            self.last_message.pop("getblocktxn", None)

    def get_headers(self, locator, hashstop):
        msg = msg_getheaders()
        msg.locator.vHave = locator
        msg.hashstop = hashstop
        self.send_message(msg)

    def send_header_for_blocks(self, new_blocks):
        headers_message = msg_headers()
        headers_message.headers = [CBlockHeader(b) for b in new_blocks]
        self.send_message(headers_message)

    def request_headers_and_sync(self, locator, hashstop=0):
        self.clear_block_announcement()
        self.get_headers(locator, hashstop)
        self.wait_until(self.received_block_announcement, timeout=30)
        self.clear_block_announcement()

    # Block until a block announcement for a particular block hash is
    # received.
    def wait_for_block_announcement(self, block_hash, timeout=30):
        def received_hash():
            return (block_hash in self.announced_blockhashes)
        self.wait_until(received_hash, timeout=timeout)

    def send_await_disconnect(self, message, timeout=30):
        """Sends a message to the node and wait for disconnect.

        This is used when we want to send a message into the node that we expect
        will get us disconnected, eg an invalid block."""
        self.send_message(message)
        self.wait_for_disconnect(timeout=timeout)

class CompactBlocksTest(BitcoinTestFramework):
    def set_test_params(self):
        self.setup_clean_chain = True
        self.num_nodes = 1
        self.extra_args = [[
            "-acceptnonstdtxn=1",
        ]]
        self.utxos = []

    def build_block_on_tip(self, node):
        block = create_block(tmpl=node.getblocktemplate(NORMAL_GBT_REQUEST_PARAMS))
        block.solve()
        return block

    # Create 10 more anyone-can-spend utxo's for testing.
    def make_utxos(self):
        block = self.build_block_on_tip(self.nodes[0])
        self.segwit_node.send_and_ping(msg_no_witness_block(block))
        assert int(self.nodes[0].getbestblockhash(), 16) == block.sha256
        self.generate(self.wallet, COINBASE_MATURITY)

        total_value = block.vtx[0].vout[0].nValue
        out_value = total_value // 10
        tx = CTransaction()
        tx.vin.append(CTxIn(COutPoint(block.vtx[0].sha256, 0), b''))
        for _ in range(10):
            tx.vout.append(CTxOut(out_value, CScript([OP_TRUE])))
        tx.rehash()

        block2 = self.build_block_on_tip(self.nodes[0])
        block2.vtx.append(tx)
        block2.hashMerkleRoot = block2.calc_merkle_root()
        block2.solve()
        self.segwit_node.send_and_ping(msg_no_witness_block(block2))
        assert_equal(int(self.nodes[0].getbestblockhash(), 16), block2.sha256)
        self.utxos.extend([[tx.sha256, i, out_value] for i in range(10)])


    # Test "sendcmpct" (between peers preferring the same version):
    # - No compact block announcements unless sendcmpct is sent.
    # - If sendcmpct is sent with version = 1, the message is ignored.
    # - If sendcmpct is sent with version > 2, the message is ignored.
    # - If sendcmpct is sent with boolean 0, then block announcements are not
    #   made with compact blocks.
    # - If sendcmpct is then sent with boolean 1, then new block announcements
    #   are made with compact blocks.
    def test_sendcmpct(self, test_node):
        node = self.nodes[0]

        # Make sure we get a SENDCMPCT message from our peer
        def received_sendcmpct():
            return (len(test_node.last_sendcmpct) > 0)
        test_node.wait_until(received_sendcmpct, timeout=30)
        with p2p_lock:
            # Check that version 2 is received.
            assert_equal(test_node.last_sendcmpct[0].version, 2)
            test_node.last_sendcmpct = []

        tip = int(node.getbestblockhash(), 16)

        def check_announcement_of_new_block(node, peer, predicate):
            peer.clear_block_announcement()
            block_hash = int(self.generate(node, 1)[0], 16)
            peer.wait_for_block_announcement(block_hash, timeout=30)
            assert peer.block_announced

            with p2p_lock:
                assert predicate(peer), (
                    "block_hash={!r}, cmpctblock={!r}, inv={!r}".format(
                        block_hash, peer.last_message.get("cmpctblock", None), peer.last_message.get("inv", None)))

        # We shouldn't get any block announcements via cmpctblock yet.
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" not in p.last_message)

        # Try one more time, this time after requesting headers.
        test_node.request_headers_and_sync(locator=[tip])
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" not in p.last_message and "inv" in p.last_message)

        # Test a few ways of using sendcmpct that should NOT
        # result in compact block announcements.
        # Before each test, sync the headers chain.
        test_node.request_headers_and_sync(locator=[tip])

        # Now try a SENDCMPCT message with too-low version
        test_node.send_and_ping(msg_sendcmpct(announce=True, version=1))
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" not in p.last_message)

        # Headers sync before next test.
        test_node.request_headers_and_sync(locator=[tip])

        # Now try a SENDCMPCT message with too-high version
        test_node.send_and_ping(msg_sendcmpct(announce=True, version=3))
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" not in p.last_message)

        # Headers sync before next test.
        test_node.request_headers_and_sync(locator=[tip])

        # Now try a SENDCMPCT message with valid version, but announce=False
        test_node.send_and_ping(msg_sendcmpct(announce=False, version=2))
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" not in p.last_message)

        # Headers sync before next test.
        test_node.request_headers_and_sync(locator=[tip])

        # Finally, try a SENDCMPCT message with announce=True
        test_node.send_and_ping(msg_sendcmpct(announce=True, version=2))
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" in p.last_message)

        # Try one more time (no headers sync should be needed!)
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" in p.last_message)

        # Try one more time, after turning on sendheaders
        test_node.send_and_ping(msg_sendheaders())
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" in p.last_message)

        # Try one more time, after sending a version=1, announce=false message.
        test_node.send_and_ping(msg_sendcmpct(announce=False, version=1))
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" in p.last_message)

        # Now turn off announcements
        test_node.send_and_ping(msg_sendcmpct(announce=False, version=2))
        check_announcement_of_new_block(node, test_node, lambda p: "cmpctblock" not in p.last_message and "headers" in p.last_message)

    # This test actually causes bitcoind to (reasonably!) disconnect us, so do this last.
    def test_invalid_cmpctblock_message(self):
        self.generate(self.nodes[0], COINBASE_MATURITY + 1)
        block = self.build_block_on_tip(self.nodes[0])

        cmpct_block = P2PHeaderAndShortIDs()
        cmpct_block.header = CBlockHeader(block)
        cmpct_block.prefilled_txn_length = 1
        # This index will be too high
        prefilled_txn = PrefilledTransaction(1, block.vtx[0])
        cmpct_block.prefilled_txn = [prefilled_txn]
        self.segwit_node.send_await_disconnect(msg_cmpctblock(cmpct_block))
        assert_equal(int(self.nodes[0].getbestblockhash(), 16), block.hashPrevBlock)

    # Compare the generated shortids to what we expect based on BIP 152, given
    # bitcoind's choice of nonce.
    def test_compactblock_construction(self, test_node):
        node = self.nodes[0]
        # Generate a bunch of transactions.
        self.generate(node, COINBASE_MATURITY + 1)
        num_transactions = 25

        segwit_tx_generated = False
        for _ in range(num_transactions):
            hex_tx = self.wallet.send_self_transfer(from_node=self.nodes[0])['hex']
            tx = tx_from_hex(hex_tx)
            if not tx.wit.is_null():
                segwit_tx_generated = True

        assert segwit_tx_generated  # check that our test is not broken

        # Wait until we've seen the block announcement for the resulting tip
        tip = int(node.getbestblockhash(), 16)
        test_node.wait_for_block_announcement(tip)

        # Make sure we will receive a fast-announce compact block
        self.request_cb_announcements(test_node)

        # Now mine a block, and look at the resulting compact block.
        test_node.clear_block_announcement()
        block_hash = int(self.generate(node, 1)[0], 16)

        # Store the raw block in our internal format.
        block = from_hex(CBlock(), node.getblock("%064x" % block_hash, False))
        for tx in block.vtx:
            tx.calc_sha256()
        block.rehash()

        # Wait until the block was announced (via compact blocks)
        test_node.wait_until(lambda: "cmpctblock" in test_node.last_message, timeout=30)

        # Now fetch and check the compact block
        header_and_shortids = None
        with p2p_lock:
            # Convert the on-the-wire representation to absolute indexes
            header_and_shortids = HeaderAndShortIDs(test_node.last_message["cmpctblock"].header_and_shortids)
        self.check_compactblock_construction_from_block(header_and_shortids, block_hash, block)

        # Now fetch the compact block using a normal non-announce getdata
        test_node.clear_block_announcement()
        inv = CInv(MSG_CMPCT_BLOCK, block_hash)
        test_node.send_message(msg_getdata([inv]))

        test_node.wait_until(lambda: "cmpctblock" in test_node.last_message, timeout=30)

        # Now fetch and check the compact block
        header_and_shortids = None
        with p2p_lock:
            # Convert the on-the-wire representation to absolute indexes
            header_and_shortids = HeaderAndShortIDs(test_node.last_message["cmpctblock"].header_and_shortids)
        self.check_compactblock_construction_from_block(header_and_shortids, block_hash, block)

    def check_compactblock_construction_from_block(self, header_and_shortids, block_hash, block):
        # Check that we got the right block!
        header_and_shortids.header.calc_sha256()
        assert_equal(header_and_shortids.header.sha256, block_hash)

        # Make sure the prefilled_txn appears to have included the coinbase
        assert len(header_and_shortids.prefilled_txn) >= 1
        assert_equal(header_and_shortids.prefilled_txn[0].index, 0)

        # Check that all prefilled_txn entries match what's in the block.
        for entry in header_and_shortids.prefilled_txn:
            entry.tx.calc_sha256()
            # This checks the non-witness parts of the tx agree
            assert_equal(entry.tx.sha256, block.vtx[entry.index].sha256)

            # And this checks the witness
            wtxid = entry.tx.calc_sha256(True)
            assert_equal(wtxid, block.vtx[entry.index].calc_sha256(True))

        # Check that the cmpctblock message announced all the transactions.
        assert_equal(len(header_and_shortids.prefilled_txn) + len(header_and_shortids.shortids), len(block.vtx))

        # And now check that all the shortids are as expected as well.
        # Determine the siphash keys to use.
        [k0, k1] = header_and_shortids.get_siphash_keys()

        index = 0
        while index < len(block.vtx):
            if (len(header_and_shortids.prefilled_txn) > 0 and
                    header_and_shortids.prefilled_txn[0].index == index):
                # Already checked prefilled transactions above
                header_and_shortids.prefilled_txn.pop(0)
            else:
                tx_hash = block.vtx[index].calc_sha256(True)
                shortid = calculate_shortid(k0, k1, tx_hash)
                assert_equal(shortid, header_and_shortids.shortids[0])
                header_and_shortids.shortids.pop(0)
            index += 1

    # Test that bitcoind requests compact blocks when we announce new blocks
    # via header or inv, and that responding to getblocktxn causes the block
    # to be successfully reconstructed.
    def test_compactblock_requests(self, test_node):
        node = self.nodes[0]
        # Try announcing a block with an inv or header, expect a compactblock
        # request
        for announce in ["inv", "header"]:
            block = self.build_block_on_tip(node)

            if announce == "inv":
                test_node.send_message(msg_inv([CInv(MSG_BLOCK, block.sha256)]))
                test_node.wait_for_getheaders(timeout=30)
                test_node.send_header_for_blocks([block])
            else:
                test_node.send_header_for_blocks([block])
            test_node.wait_for_getdata([block.sha256], timeout=30)
            assert_equal(test_node.last_message["getdata"].inv[0].type, 4)

            # Send back a compactblock message that omits the coinbase
            comp_block = HeaderAndShortIDs()
            comp_block.header = CBlockHeader(block)
            comp_block.nonce = 0
            [k0, k1] = comp_block.get_siphash_keys()
            coinbase_hash = block.vtx[0].calc_sha256(True)
            comp_block.shortids = [calculate_shortid(k0, k1, coinbase_hash)]
            test_node.send_and_ping(msg_cmpctblock(comp_block.to_p2p()))
            assert_equal(int(node.getbestblockhash(), 16), block.hashPrevBlock)
            # Expect a getblocktxn message.
            with p2p_lock:
                assert "getblocktxn" in test_node.last_message
                absolute_indexes = test_node.last_message["getblocktxn"].block_txn_request.to_absolute()
            assert_equal(absolute_indexes, [0])  # should be a coinbase request

            # Send the coinbase, and verify that the tip advances.
            msg = msg_blocktxn()
            msg.block_transactions.blockhash = block.sha256
            msg.block_transactions.transactions = [block.vtx[0]]
            test_node.send_and_ping(msg)
            assert_equal(int(node.getbestblockhash(), 16), block.sha256)

    # Create a chain of transactions from given utxo, and add to a new block.
    def build_block_with_transactions(self, node, utxo, num_transactions):
        block = self.build_block_on_tip(node)

        for _ in range(num_transactions):
            tx = CTransaction()
            tx.vin.append(CTxIn(COutPoint(utxo[0], utxo[1]), b''))
            tx.vout.append(CTxOut(utxo[2] - 1000, CScript([OP_TRUE, OP_DROP] * 15 + [OP_TRUE])))
            tx.rehash()
            utxo = [tx.sha256, 0, tx.vout[0].nValue]
            block.vtx.append(tx)

        block.hashMerkleRoot = block.calc_merkle_root()
        block.solve()
        return block

    # Test that we only receive getblocktxn requests for transactions that the
    # node needs, and that responding to them causes the block to be
    # reconstructed.
    def test_getblocktxn_requests(self, test_node):
        node = self.nodes[0]

        def test_getblocktxn_response(compact_block, peer, expected_result):
            msg = msg_cmpctblock(compact_block.to_p2p())
            peer.send_and_ping(msg)
            with p2p_lock:
                assert "getblocktxn" in peer.last_message
                absolute_indexes = peer.last_message["getblocktxn"].block_txn_request.to_absolute()
            assert_equal(absolute_indexes, expected_result)

        def test_tip_after_message(node, peer, msg, tip):
            peer.send_and_ping(msg)
            assert_equal(int(node.getbestblockhash(), 16), tip)

        # First try announcing compactblocks that won't reconstruct, and verify
        # that we receive getblocktxn messages back.
        utxo = self.utxos.pop(0)

        block = self.build_block_with_transactions(node, utxo, 5)
        self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])
        comp_block = HeaderAndShortIDs()
        comp_block.initialize_from_block(block, use_witness=True)

        test_getblocktxn_response(comp_block, test_node, [1, 2, 3, 4, 5])

        msg_bt = msg_no_witness_blocktxn()
        msg_bt = msg_blocktxn()  # serialize with witnesses
        msg_bt.block_transactions = BlockTransactions(block.sha256, block.vtx[1:])
        test_tip_after_message(node, test_node, msg_bt, block.sha256)

        utxo = self.utxos.pop(0)
        block = self.build_block_with_transactions(node, utxo, 5)
        self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])

        # Now try interspersing the prefilled transactions
        comp_block.initialize_from_block(block, prefill_list=[0, 1, 5], use_witness=True)
        test_getblocktxn_response(comp_block, test_node, [2, 3, 4])
        msg_bt.block_transactions = BlockTransactions(block.sha256, block.vtx[2:5])
        test_tip_after_message(node, test_node, msg_bt, block.sha256)

        # Now try giving one transaction ahead of time.
        utxo = self.utxos.pop(0)
        block = self.build_block_with_transactions(node, utxo, 5)
        self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])
        test_node.send_and_ping(msg_tx(block.vtx[1]))
        assert block.vtx[1].hash in node.getrawmempool()

        # Prefill 4 out of the 6 transactions, and verify that only the one
        # that was not in the mempool is requested.
        comp_block.initialize_from_block(block, prefill_list=[0, 2, 3, 4], use_witness=True)
        test_getblocktxn_response(comp_block, test_node, [5])

        msg_bt.block_transactions = BlockTransactions(block.sha256, [block.vtx[5]])
        test_tip_after_message(node, test_node, msg_bt, block.sha256)

        # Now provide all transactions to the node before the block is
        # announced and verify reconstruction happens immediately.
        utxo = self.utxos.pop(0)
        block = self.build_block_with_transactions(node, utxo, 10)
        self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])
        for tx in block.vtx[1:]:
            test_node.send_message(msg_tx(tx))
        test_node.sync_with_ping()
        # Make sure all transactions were accepted.
        mempool = node.getrawmempool()
        for tx in block.vtx[1:]:
            assert tx.hash in mempool

        # Clear out last request.
        with p2p_lock:
            test_node.last_message.pop("getblocktxn", None)

        # Send compact block
        comp_block.initialize_from_block(block, prefill_list=[0], use_witness=True)
        test_tip_after_message(node, test_node, msg_cmpctblock(comp_block.to_p2p()), block.sha256)
        with p2p_lock:
            # Shouldn't have gotten a request for any transaction
            assert "getblocktxn" not in test_node.last_message

    # Incorrectly responding to a getblocktxn shouldn't cause the block to be
    # permanently failed.
    def test_incorrect_blocktxn_response(self, test_node):
        node = self.nodes[0]
        utxo = self.utxos.pop(0)

        block = self.build_block_with_transactions(node, utxo, 10)
        self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])
        # Relay the first 5 transactions from the block in advance
        for tx in block.vtx[1:6]:
            test_node.send_message(msg_tx(tx))
        test_node.sync_with_ping()
        # Make sure all transactions were accepted.
        mempool = node.getrawmempool()
        for tx in block.vtx[1:6]:
            assert tx.hash in mempool

        # Send compact block
        comp_block = HeaderAndShortIDs()
        comp_block.initialize_from_block(block, prefill_list=[0], use_witness=True)
        test_node.send_and_ping(msg_cmpctblock(comp_block.to_p2p()))
        absolute_indexes = []
        with p2p_lock:
            assert "getblocktxn" in test_node.last_message
            absolute_indexes = test_node.last_message["getblocktxn"].block_txn_request.to_absolute()
        assert_equal(absolute_indexes, [6, 7, 8, 9, 10])

        # Now give an incorrect response.
        # Note that it's possible for bitcoind to be smart enough to know we're
        # lying, since it could check to see if the shortid matches what we're
        # sending, and eg disconnect us for misbehavior.  If that behavior
        # change was made, we could just modify this test by having a
        # different peer provide the block further down, so that we're still
        # verifying that the block isn't marked bad permanently. This is good
        # enough for now.
        msg = msg_blocktxn()
        msg.block_transactions = BlockTransactions(block.sha256, [block.vtx[5]] + block.vtx[7:])
        test_node.send_and_ping(msg)

        # Tip should not have updated
        assert_equal(int(node.getbestblockhash(), 16), block.hashPrevBlock)

        # We should receive a getdata request
        test_node.wait_for_getdata([block.sha256], timeout=10)
        assert test_node.last_message["getdata"].inv[0].type == MSG_BLOCK or \
               test_node.last_message["getdata"].inv[0].type == MSG_BLOCK | MSG_WITNESS_FLAG

        # Deliver the block
        test_node.send_and_ping(msg_block(block))
        assert_equal(int(node.getbestblockhash(), 16), block.sha256)

    # Multiple blocktxn responses will cause a node to get disconnected.
    def test_multiple_blocktxn_response(self, test_node):
        node = self.nodes[0]
        utxo = self.utxos[0]

        block = self.build_block_with_transactions(node, utxo, 2)

        # Send compact block
        comp_block = HeaderAndShortIDs()
        comp_block.initialize_from_block(block, prefill_list=[0], use_witness=True)
        test_node.send_and_ping(msg_cmpctblock(comp_block.to_p2p()))
        absolute_indexes = []
        with p2p_lock:
            assert "getblocktxn" in test_node.last_message
            absolute_indexes = test_node.last_message["getblocktxn"].block_txn_request.to_absolute()
        assert_equal(absolute_indexes, [1, 2])

        # Send a blocktxn that does not succeed in reconstruction, triggering
        # getdata fallback.
        msg = msg_blocktxn()
        msg.block_transactions = BlockTransactions(block.sha256, [block.vtx[2]] + [block.vtx[1]])
        test_node.send_and_ping(msg)

        # Tip should not have updated
        assert_equal(int(node.getbestblockhash(), 16), block.hashPrevBlock)

        # We should receive a getdata request
        test_node.wait_for_getdata([block.sha256], timeout=10)
        assert test_node.last_message["getdata"].inv[0].type == MSG_BLOCK or \
               test_node.last_message["getdata"].inv[0].type == MSG_BLOCK | MSG_WITNESS_FLAG

        # Send the same blocktxn and assert the sender gets disconnected.
        with node.assert_debug_log(['previous compact block reconstruction attempt failed']):
            test_node.send_message(msg)
        test_node.wait_for_disconnect()

    def test_getblocktxn_handler(self, test_node):
        node = self.nodes[0]
        # bitcoind will not send blocktxn responses for blocks whose height is
        # more than 10 blocks deep.
        MAX_GETBLOCKTXN_DEPTH = 10
        chain_height = node.getblockcount()
        current_height = chain_height
        while (current_height >= chain_height - MAX_GETBLOCKTXN_DEPTH):
            block_hash = node.getblockhash(current_height)
            block = from_hex(CBlock(), node.getblock(block_hash, False))

            msg = msg_getblocktxn()
            msg.block_txn_request = BlockTransactionsRequest(int(block_hash, 16), [])
            num_to_request = random.randint(1, len(block.vtx))
            msg.block_txn_request.from_absolute(sorted(random.sample(range(len(block.vtx)), num_to_request)))
            test_node.send_message(msg)
            test_node.wait_until(lambda: "blocktxn" in test_node.last_message, timeout=10)

            [tx.calc_sha256() for tx in block.vtx]
            with p2p_lock:
                assert_equal(test_node.last_message["blocktxn"].block_transactions.blockhash, int(block_hash, 16))
                all_indices = msg.block_txn_request.to_absolute()
                for index in all_indices:
                    tx = test_node.last_message["blocktxn"].block_transactions.transactions.pop(0)
                    tx.calc_sha256()
                    assert_equal(tx.sha256, block.vtx[index].sha256)
                    # Check that the witness matches
                    assert_equal(tx.calc_sha256(True), block.vtx[index].calc_sha256(True))
                test_node.last_message.pop("blocktxn", None)
            current_height -= 1

        # Next request should send a full block response, as we're past the
        # allowed depth for a blocktxn response.
        block_hash = node.getblockhash(current_height)
        msg.block_txn_request = BlockTransactionsRequest(int(block_hash, 16), [0])
        with p2p_lock:
            test_node.last_message.pop("block", None)
            test_node.last_message.pop("blocktxn", None)
        test_node.send_and_ping(msg)
        with p2p_lock:
            test_node.last_message["block"].block.calc_sha256()
            assert_equal(test_node.last_message["block"].block.sha256, int(block_hash, 16))
            assert "blocktxn" not in test_node.last_message

        # Request with out-of-bounds tx index results in disconnect
        bad_peer = self.nodes[0].add_p2p_connection(TestP2PConn())
        block_hash = node.getblockhash(chain_height)
        block = from_hex(CBlock(), node.getblock(block_hash, False))
        msg.block_txn_request = BlockTransactionsRequest(int(block_hash, 16), [len(block.vtx)])
        with node.assert_debug_log(['getblocktxn with out-of-bounds tx indices']):
            bad_peer.send_message(msg)
        bad_peer.wait_for_disconnect()

    def test_low_work_compactblocks(self, test_node):
        # A compactblock with insufficient work won't get its header included
        node = self.nodes[0]
        hashPrevBlock = int(node.getblockhash(node.getblockcount() - 150), 16)
        block = self.build_block_on_tip(node)
        block.hashPrevBlock = hashPrevBlock
        block.solve()

        comp_block = HeaderAndShortIDs()
        comp_block.initialize_from_block(block)
        with self.nodes[0].assert_debug_log(['[net] Ignoring low-work compact block from peer 0']):
            test_node.send_and_ping(msg_cmpctblock(comp_block.to_p2p()))

        tips = node.getchaintips()
        found = False
        for x in tips:
            if x["hash"] == block.hash:
                found = True
                break
        assert not found

    def test_compactblocks_not_at_tip(self, test_node):
        node = self.nodes[0]
        # Test that requesting old compactblocks doesn't work.
        MAX_CMPCTBLOCK_DEPTH = 5
        new_blocks = []
        for _ in range(MAX_CMPCTBLOCK_DEPTH + 1):
            test_node.clear_block_announcement()
            new_blocks.append(self.generate(node, 1)[0])
            test_node.wait_until(test_node.received_block_announcement, timeout=30)

        test_node.clear_block_announcement()
        test_node.send_message(msg_getdata([CInv(MSG_CMPCT_BLOCK, int(new_blocks[0], 16))]))
        test_node.wait_until(lambda: "cmpctblock" in test_node.last_message, timeout=30)

        test_node.clear_block_announcement()
        self.generate(node, 1)
        test_node.wait_until(test_node.received_block_announcement, timeout=30)
        test_node.clear_block_announcement()
        with p2p_lock:
            test_node.last_message.pop("block", None)
        test_node.send_message(msg_getdata([CInv(MSG_CMPCT_BLOCK, int(new_blocks[0], 16))]))
        test_node.wait_until(lambda: "block" in test_node.last_message, timeout=30)
        with p2p_lock:
            test_node.last_message["block"].block.calc_sha256()
            assert_equal(test_node.last_message["block"].block.sha256, int(new_blocks[0], 16))

        # Generate an old compactblock, and verify that it's not accepted.
        cur_height = node.getblockcount()
        hashPrevBlock = int(node.getblockhash(cur_height - 5), 16)
        block = self.build_block_on_tip(node)
        block.hashPrevBlock = hashPrevBlock
        block.solve()

        comp_block = HeaderAndShortIDs()
        comp_block.initialize_from_block(block)
        test_node.send_and_ping(msg_cmpctblock(comp_block.to_p2p()))

        tips = node.getchaintips()
        found = False
        for x in tips:
            if x["hash"] == block.hash:
                assert_equal(x["status"], "headers-only")
                found = True
                break
        assert found

        # Requesting this block via getblocktxn should silently fail
        # (to avoid fingerprinting attacks).
        msg = msg_getblocktxn()
        msg.block_txn_request = BlockTransactionsRequest(block.sha256, [0])
        with p2p_lock:
            test_node.last_message.pop("blocktxn", None)
        test_node.send_and_ping(msg)
        with p2p_lock:
            assert "blocktxn" not in test_node.last_message

    def test_end_to_end_block_relay(self, listeners):
        node = self.nodes[0]
        utxo = self.utxos.pop(0)

        block = self.build_block_with_transactions(node, utxo, 10)

        [l.clear_block_announcement() for l in listeners]

        # serialize without witness (this block has no witnesses anyway).
        # TODO: repeat this test with witness tx's to a segwit node.
        node.submitblock(block.serialize().hex())

        for l in listeners:
            l.wait_until(lambda: "cmpctblock" in l.last_message, timeout=30)
        with p2p_lock:
            for l in listeners:
                l.last_message["cmpctblock"].header_and_shortids.header.calc_sha256()
                assert_equal(l.last_message["cmpctblock"].header_and_shortids.header.sha256, block.sha256)

    # Test that we don't get disconnected if we relay a compact block with valid header,
    # but invalid transactions.
    def test_invalid_tx_in_compactblock(self, test_node):
        node = self.nodes[0]
        assert len(self.utxos)
        utxo = self.utxos[0]

        block = self.build_block_with_transactions(node, utxo, 5)
        del block.vtx[3]
        block.hashMerkleRoot = block.calc_merkle_root()
        # Drop the coinbase witness but include the witness commitment.
        add_witness_commitment(block)
        block.vtx[0].wit.vtxinwit = []
        block.solve()

        # Now send the compact block with all transactions prefilled, and
        # verify that we don't get disconnected.
        comp_block = HeaderAndShortIDs()
        comp_block.initialize_from_block(block, prefill_list=list(range(len(block.vtx))), use_witness=True)
        msg = msg_cmpctblock(comp_block.to_p2p())
        test_node.send_and_ping(msg)

        # Check that the tip didn't advance
        assert int(node.getbestblockhash(), 16) != block.sha256
        test_node.sync_with_ping()

    # Helper for enabling cb announcements
    # Send the sendcmpct request and sync headers
    def request_cb_announcements(self, peer):
        node = self.nodes[0]
        tip = node.getbestblockhash()
        peer.get_headers(locator=[int(tip, 16)], hashstop=0)
        peer.send_and_ping(msg_sendcmpct(announce=True, version=2))

    def test_compactblock_reconstruction_stalling_peer(self, stalling_peer, delivery_peer):
        node = self.nodes[0]
        assert len(self.utxos)

        def announce_cmpct_block(node, peer):
            utxo = self.utxos.pop(0)
            block = self.build_block_with_transactions(node, utxo, 5)

            cmpct_block = HeaderAndShortIDs()
            cmpct_block.initialize_from_block(block)
            msg = msg_cmpctblock(cmpct_block.to_p2p())
            peer.send_and_ping(msg)
            with p2p_lock:
                assert "getblocktxn" in peer.last_message
            return block, cmpct_block

        block, cmpct_block = announce_cmpct_block(node, stalling_peer)

        for tx in block.vtx[1:]:
            delivery_peer.send_message(msg_tx(tx))
        delivery_peer.sync_with_ping()
        mempool = node.getrawmempool()
        for tx in block.vtx[1:]:
            assert tx.hash in mempool

        delivery_peer.send_and_ping(msg_cmpctblock(cmpct_block.to_p2p()))
        assert_equal(int(node.getbestblockhash(), 16), block.sha256)

        self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])

        # Now test that delivering an invalid compact block won't break relay

        block, cmpct_block = announce_cmpct_block(node, stalling_peer)
        for tx in block.vtx[1:]:
            delivery_peer.send_message(msg_tx(tx))
        delivery_peer.sync_with_ping()

        cmpct_block.prefilled_txn[0].tx.wit.vtxinwit = [CTxInWitness()]
        cmpct_block.prefilled_txn[0].tx.wit.vtxinwit[0].scriptWitness.stack = [ser_uint256(0)]

        cmpct_block.use_witness = True
        delivery_peer.send_and_ping(msg_cmpctblock(cmpct_block.to_p2p()))
        assert int(node.getbestblockhash(), 16) != block.sha256

        msg = msg_no_witness_blocktxn()
        msg.block_transactions.blockhash = block.sha256
        msg.block_transactions.transactions = block.vtx[1:]
        stalling_peer.send_and_ping(msg)
        assert_equal(int(node.getbestblockhash(), 16), block.sha256)

    def test_highbandwidth_mode_states_via_getpeerinfo(self):
        # create new p2p connection for a fresh state w/o any prior sendcmpct messages sent
        hb_test_node = self.nodes[0].add_p2p_connection(TestP2PConn())

        # assert the RPC getpeerinfo boolean fields `bip152_hb_{to, from}`
        # match the given parameters for the last peer of a given node
        def assert_highbandwidth_states(node, hb_to, hb_from):
            peerinfo = node.getpeerinfo()[-1]
            assert_equal(peerinfo['bip152_hb_to'], hb_to)
            assert_equal(peerinfo['bip152_hb_from'], hb_from)

        # initially, neither node has selected the other peer as high-bandwidth yet
        assert_highbandwidth_states(self.nodes[0], hb_to=False, hb_from=False)

        # peer requests high-bandwidth mode by sending sendcmpct(1)
        hb_test_node.send_and_ping(msg_sendcmpct(announce=True, version=2))
        assert_highbandwidth_states(self.nodes[0], hb_to=False, hb_from=True)

        # peer generates a block and sends it to node, which should
        # select the peer as high-bandwidth (up to 3 peers according to BIP 152)
        block = self.build_block_on_tip(self.nodes[0])
        hb_test_node.send_and_ping(msg_block(block))
        assert_highbandwidth_states(self.nodes[0], hb_to=True, hb_from=True)

        # peer requests low-bandwidth mode by sending sendcmpct(0)
        hb_test_node.send_and_ping(msg_sendcmpct(announce=False, version=2))
        assert_highbandwidth_states(self.nodes[0], hb_to=True, hb_from=False)

    def test_compactblock_reconstruction_parallel_reconstruction(self, stalling_peer, delivery_peer, inbound_peer, outbound_peer):
        """ All p2p connections are inbound except outbound_peer. We test that ultimate parallel slot
            can only be taken by an outbound node unless prior attempts were done by an outbound
        """
        node = self.nodes[0]
        assert len(self.utxos)

        def announce_cmpct_block(node, peer, txn_count):
            utxo = self.utxos.pop(0)
            block = self.build_block_with_transactions(node, utxo, txn_count)

            cmpct_block = HeaderAndShortIDs()
            cmpct_block.initialize_from_block(block)
            msg = msg_cmpctblock(cmpct_block.to_p2p())
            peer.send_and_ping(msg)
            with p2p_lock:
                assert "getblocktxn" in peer.last_message
            return block, cmpct_block

        for name, peer in [("delivery", delivery_peer), ("inbound", inbound_peer), ("outbound", outbound_peer)]:
            self.log.info(f"Setting {name} as high bandwidth peer")
            block, cmpct_block = announce_cmpct_block(node, peer, 1)
            msg = msg_blocktxn()
            msg.block_transactions.blockhash = block.sha256
            msg.block_transactions.transactions = block.vtx[1:]
            peer.send_and_ping(msg)
            assert_equal(int(node.getbestblockhash(), 16), block.sha256)
            peer.clear_getblocktxn()

        # Test the simple parallel download case...
        for num_missing in [1, 5, 20]:

            # Remaining low-bandwidth peer is stalling_peer, who announces first
            assert_equal([peer['bip152_hb_to'] for peer in node.getpeerinfo()], [False, True, True, True])

            block, cmpct_block = announce_cmpct_block(node, stalling_peer, num_missing)

            delivery_peer.send_and_ping(msg_cmpctblock(cmpct_block.to_p2p()))
            with p2p_lock:
                # The second peer to announce should still get a getblocktxn
                assert "getblocktxn" in delivery_peer.last_message
            assert int(node.getbestblockhash(), 16) != block.sha256

            inbound_peer.send_and_ping(msg_cmpctblock(cmpct_block.to_p2p()))
            with p2p_lock:
                # The third inbound peer to announce should *not* get a getblocktxn
                assert "getblocktxn" not in inbound_peer.last_message
            assert int(node.getbestblockhash(), 16) != block.sha256

            outbound_peer.send_and_ping(msg_cmpctblock(cmpct_block.to_p2p()))
            with p2p_lock:
                # The third peer to announce should get a getblocktxn if outbound
                assert "getblocktxn" in outbound_peer.last_message
            assert int(node.getbestblockhash(), 16) != block.sha256

            # Second peer completes the compact block first
            msg = msg_blocktxn()
            msg.block_transactions.blockhash = block.sha256
            msg.block_transactions.transactions = block.vtx[1:]
            delivery_peer.send_and_ping(msg)
            assert_equal(int(node.getbestblockhash(), 16), block.sha256)

            # Nothing bad should happen if we get a late fill from the first peer...
            stalling_peer.send_and_ping(msg)
            self.utxos.append([block.vtx[-1].sha256, 0, block.vtx[-1].vout[0].nValue])

            delivery_peer.clear_getblocktxn()
            inbound_peer.clear_getblocktxn()
            outbound_peer.clear_getblocktxn()


    def run_test(self):
        self.wallet = MiniWallet(self.nodes[0])

        # Setup the p2p connections
        self.segwit_node = self.nodes[0].add_p2p_connection(TestP2PConn())
        self.additional_segwit_node = self.nodes[0].add_p2p_connection(TestP2PConn())
        self.onemore_inbound_node = self.nodes[0].add_p2p_connection(TestP2PConn())
        self.outbound_node = self.nodes[0].add_outbound_p2p_connection(TestP2PConn(), p2p_idx=3, connection_type="outbound-full-relay")

        # We will need UTXOs to construct transactions in later tests.
        self.make_utxos()

        assert softfork_active(self.nodes[0], "segwit")

        self.log.info("Testing SENDCMPCT p2p message... ")
        self.test_sendcmpct(self.segwit_node)
        self.test_sendcmpct(self.additional_segwit_node)
        self.test_sendcmpct(self.onemore_inbound_node)
        self.test_sendcmpct(self.outbound_node)

        self.log.info("Testing compactblock construction...")
        self.test_compactblock_construction(self.segwit_node)

        self.log.info("Testing compactblock requests (segwit node)... ")
        self.test_compactblock_requests(self.segwit_node)

        self.log.info("Testing getblocktxn requests (segwit node)...")
        self.test_getblocktxn_requests(self.segwit_node)

        self.log.info("Testing getblocktxn handler (segwit node should return witnesses)...")
        self.test_getblocktxn_handler(self.segwit_node)

        self.log.info("Testing compactblock requests/announcements not at chain tip...")
        self.test_compactblocks_not_at_tip(self.segwit_node)

        self.log.info("Testing handling of low-work compact blocks...")
        self.test_low_work_compactblocks(self.segwit_node)

        self.log.info("Testing handling of incorrect blocktxn responses...")
        self.test_incorrect_blocktxn_response(self.segwit_node)

        self.log.info("Testing reconstructing compact blocks with a stalling peer...")
        self.test_compactblock_reconstruction_stalling_peer(self.segwit_node, self.additional_segwit_node)

        self.log.info("Testing reconstructing compact blocks from multiple peers...")
        self.test_compactblock_reconstruction_parallel_reconstruction(stalling_peer=self.segwit_node, inbound_peer=self.onemore_inbound_node, delivery_peer=self.additional_segwit_node, outbound_peer=self.outbound_node)

        # Test that if we submitblock to node1, we'll get a compact block
        # announcement to all peers.
        # (Post-segwit activation, blocks won't propagate from node0 to node1
        # automatically, so don't bother testing a block announced to node0.)
        self.log.info("Testing end-to-end block relay...")
        self.request_cb_announcements(self.segwit_node)
        self.request_cb_announcements(self.additional_segwit_node)
        self.test_end_to_end_block_relay([self.segwit_node, self.additional_segwit_node])

        self.log.info("Testing handling of invalid compact blocks...")
        self.test_invalid_tx_in_compactblock(self.segwit_node)

        self.log.info("Testing handling of multiple blocktxn responses...")
        self.test_multiple_blocktxn_response(self.segwit_node)

        # The previous test will lead to a disconnection. Reconnect before continuing.
        self.segwit_node = self.nodes[0].add_p2p_connection(TestP2PConn())

        self.log.info("Testing invalid index in cmpctblock message...")
        self.test_invalid_cmpctblock_message()

        self.log.info("Testing high-bandwidth mode states via getpeerinfo...")
        self.test_highbandwidth_mode_states_via_getpeerinfo()


if __name__ == '__main__':
    CompactBlocksTest(__file__).main()



#!/usr/bin/env python3
# Copyright (c) 2015-2022 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Test BIP66 (DER SIG).

Test the DERSIG soft-fork activation on regtest.
"""

from test_framework.blocktools import (
    create_block,
    create_coinbase,
)
from test_framework.messages import msg_block
from test_framework.p2p import P2PInterface
from test_framework.script import CScript
from test_framework.test_framework import BitcoinTestFramework
from test_framework.util import (
    assert_equal,
)
from test_framework.wallet import (
    MiniWallet,
    MiniWalletMode,
)


# A canonical signature consists of:
# <30> <total len> <02> <len R> <R> <02> <len S> <S> <hashtype>
def unDERify(tx):
    """
    Make the signature in vin 0 of a tx non-DER-compliant,
    by adding padding after the S-value.
    """
    scriptSig = CScript(tx.vin[0].scriptSig)
    newscript = []
    for i in scriptSig:
        if (len(newscript) == 0):
            newscript.append(i[0:-1] + b'\0' + i[-1:])
        else:
            newscript.append(i)
    tx.vin[0].scriptSig = CScript(newscript)


DERSIG_HEIGHT = 102


class BIP66Test(BitcoinTestFramework):
    def set_test_params(self):
        self.num_nodes = 1
        # whitelist peers to speed up tx relay / mempool sync
        self.noban_tx_relay = True
        self.extra_args = [[
            f'-testactivationheight=dersig@{DERSIG_HEIGHT}',
        ]]
        self.setup_clean_chain = True
        self.rpc_timeout = 240

    def create_tx(self, input_txid):
        utxo_to_spend = self.miniwallet.get_utxo(txid=input_txid, mark_as_spent=False)
        return self.miniwallet.create_self_transfer(utxo_to_spend=utxo_to_spend)['tx']

    def test_dersig_info(self, *, is_active):
        assert_equal(self.nodes[0].getdeploymentinfo()['deployments']['bip66'],
            {
                "active": is_active,
                "height": DERSIG_HEIGHT,
                "type": "buried",
            },
        )

    def run_test(self):
        peer = self.nodes[0].add_p2p_connection(P2PInterface())
        self.miniwallet = MiniWallet(self.nodes[0], mode=MiniWalletMode.RAW_P2PK)

        self.test_dersig_info(is_active=False)

        self.log.info("Mining %d blocks", DERSIG_HEIGHT - 2)
        self.coinbase_txids = [self.nodes[0].getblock(b)['tx'][0] for b in self.generate(self.miniwallet, DERSIG_HEIGHT - 2)]

        self.log.info("Test that a transaction with non-DER signature can still appear in a block")

        spendtx = self.create_tx(self.coinbase_txids[0])
        unDERify(spendtx)
        spendtx.rehash()

        tip = self.nodes[0].getbestblockhash()
        block_time = self.nodes[0].getblockheader(tip)['mediantime'] + 1
        block = create_block(int(tip, 16), create_coinbase(DERSIG_HEIGHT - 1), block_time, txlist=[spendtx])
        block.solve()

        assert_equal(self.nodes[0].getblockcount(), DERSIG_HEIGHT - 2)
        self.test_dersig_info(is_active=False)  # Not active as of current tip and next block does not need to obey rules
        peer.send_and_ping(msg_block(block))
        assert_equal(self.nodes[0].getblockcount(), DERSIG_HEIGHT - 1)
        self.test_dersig_info(is_active=True)  # Not active as of current tip, but next block must obey rules
        assert_equal(self.nodes[0].getbestblockhash(), block.hash)

        self.log.info("Test that blocks must now be at least version 3")
        tip = block.sha256
        block_time += 1
        block = create_block(tip, create_coinbase(DERSIG_HEIGHT), block_time, version=2)
        block.solve()

        with self.nodes[0].assert_debug_log(expected_msgs=[f'{block.hash}, bad-version(0x00000002)']):
            peer.send_and_ping(msg_block(block))
            assert_equal(int(self.nodes[0].getbestblockhash(), 16), tip)
            peer.sync_with_ping()

        self.log.info("Test that transactions with non-DER signatures cannot appear in a block")
        block.nVersion = 4

        coin_txid = self.coinbase_txids[1]
        spendtx = self.create_tx(coin_txid)
        unDERify(spendtx)
        spendtx.rehash()

        # First we show that this tx is valid except for DERSIG by getting it
        # rejected from the mempool for exactly that reason.
        spendtx_txid = spendtx.hash
        spendtx_wtxid = spendtx.getwtxid()
        expected = {
                'txid': spendtx_txid,
                'wtxid': spendtx_wtxid,
                'allowed': False,
                'reject-reason': 'mempool-script-verify-flag-failed (Non-canonical DER signature)',
                'reject-details': 'mempool-script-verify-flag-failed (Non-canonical DER signature), ' +
                                  f"input 0 of {spendtx_txid} (wtxid {spendtx_wtxid}), spending {coin_txid}:0",
        }
        result = self.nodes[0].testmempoolaccept(rawtxs=[spendtx.serialize().hex()], maxfeerate=0)[0]
        # skip for now
        result.pop('usage')
        assert_equal(result, expected)

        # Now we verify that a block with this transaction is also invalid.
        block.vtx.append(spendtx)
        block.hashMerkleRoot = block.calc_merkle_root()
        block.solve()

        with self.nodes[0].assert_debug_log(expected_msgs=['Block validation error: mandatory-script-verify-flag-failed (Non-canonical DER signature)']):
            peer.send_and_ping(msg_block(block))
            assert_equal(int(self.nodes[0].getbestblockhash(), 16), tip)
            peer.sync_with_ping()

        self.log.info("Test that a block with a DERSIG-compliant transaction is accepted")
        block.vtx[1] = self.create_tx(self.coinbase_txids[1])
        block.hashMerkleRoot = block.calc_merkle_root()
        block.solve()

        self.test_dersig_info(is_active=True)  # Not active as of current tip, but next block must obey rules
        peer.send_and_ping(msg_block(block))
        self.test_dersig_info(is_active=True)  # Active as of current tip
        assert_equal(int(self.nodes[0].getbestblockhash(), 16), block.sha256)


if __name__ == '__main__':
    BIP66Test(__file__).main()



#!/usr/bin/env python3
# Copyright (c) 2025 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Test ReducedData Temporary Softfork (RDTS) consensus rules.

This test verifies all 7 consensus rules enforced by DEPLOYMENT_REDUCED_DATA:

1. Output scriptPubKeys exceeding 34 bytes are invalid (except OP_RETURN up to 83 bytes)
2. OP_PUSHDATA* with payloads larger than 256 bytes are invalid (except BIP16 redeemScript)
3. Spending undefined witness versions (not v0/v1) is invalid
4. Witness stacks with a Taproot annex are invalid
5. Taproot control blocks larger than 257 bytes are invalid (max 7 merkle nodes = 128 leaves)
6. Tapscripts including OP_SUCCESS* opcodes are invalid
7. Tapscripts executing OP_IF or OP_NOTIF instructions are invalid
"""

from test_framework.test_framework import BitcoinTestFramework
from test_framework.wallet import MiniWallet
from test_framework.messages import (
    CBlock,
    COutPoint,
    CTransaction,
    CTxIn,
    CTxInWitness,
    CTxOut,
    COIN,
    MAX_OP_RETURN_RELAY,
)
from test_framework.p2p import P2PDataStore
from test_framework.script import (
    ANNEX_TAG,
    CScript,
    CScriptOp,
    is_op_success,
    LEAF_VERSION_TAPSCRIPT,
    OP_0,
    OP_1,
    OP_2,
    OP_3,
    OP_4,
    OP_5,
    OP_6,
    OP_7,
    OP_8,
    OP_9,
    OP_10,
    OP_11,
    OP_12,
    OP_13,
    OP_14,
    OP_15,
    OP_16,
    OP_CHECKSIG,
    OP_CHECKSIGADD,
    OP_CHECKMULTISIG,
    OP_DROP,
    OP_DUP,
    OP_EQUAL,
    OP_EQUALVERIFY,
    OP_HASH160,
    OP_IF,
    OP_NOTIF,
    OP_ENDIF,
    OP_PUSHDATA1,
    OP_PUSHDATA2,
    OP_RETURN,
    OP_TRUE,
    SegwitV0SignatureHash,
    SIGHASH_ALL,
    SIGHASH_DEFAULT,
    hash160,
    sha256,
    taproot_construct,
    TaprootSignatureHash,
)
from test_framework.blocktools import (
    create_block,
    create_coinbase,
    add_witness_commitment,
)
from test_framework.script_util import (
    PAY_TO_ANCHOR,
    script_to_p2wsh_script,
    script_to_p2sh_script,
)
from test_framework.util import (
    assert_equal,
    assert_raises_rpc_error,
)
from test_framework.key import (
    ECKey,
    compute_xonly_pubkey,
    generate_privkey,
    sign_schnorr,
    tweak_add_privkey,
)
from io import BytesIO
import struct


# Constants from BIP444
MAX_OUTPUT_SCRIPT_SIZE = 34
MAX_OUTPUT_DATA_SIZE = 83
MAX_SCRIPT_ELEMENT_SIZE_REDUCED = 256
TAPROOT_CONTROL_BASE_SIZE = 33
TAPROOT_CONTROL_NODE_SIZE = 32
TAPROOT_CONTROL_MAX_NODE_COUNT_REDUCED = 7
TAPROOT_CONTROL_MAX_SIZE_REDUCED = TAPROOT_CONTROL_BASE_SIZE + TAPROOT_CONTROL_NODE_SIZE * TAPROOT_CONTROL_MAX_NODE_COUNT_REDUCED
# ANNEX_TAG is imported from test_framework.script


class ReducedDataTest(BitcoinTestFramework):
    def set_test_params(self):
        self.num_nodes = 1
        self.setup_clean_chain = True
        # Make DEPLOYMENT_REDUCED_DATA always active (from block 0)
        # Using start_time=-1 (ALWAYS_ACTIVE) bypasses BIP9 state machine
        self.extra_args = [[
            '-vbparams=reduced_data:-1:999999999999:0',
            '-acceptnonstdtxn=1',
        ]]

    def init_test(self):
        """Initialize test by mining blocks and creating UTXOs."""
        node = self.nodes[0]

        # MiniWallet provides a simple wallet for test transactions
        self.wallet = MiniWallet(node)

        # Mine 120 blocks to mature coinbase outputs and create spending UTXOs
        # (101 for maturity + extras since each test consumes a UTXO)
        self.generate(self.wallet, 120)

        self.log.info("Test initialization complete")

    def create_test_transaction(self, scriptPubKey, value=None):
        """Helper to create a transaction with custom scriptPubKey (not broadcast)."""
        # Start with a valid transaction from the wallet
        tx_dict = self.wallet.create_self_transfer()
        tx = tx_dict['tx']

        # Use default output value if not specified (handles fee calculation)
        if value is None:
            value = tx.vout[0].nValue

        # Replace output with our custom scriptPubKey
        tx.vout[0] = CTxOut(value, scriptPubKey)
        tx.rehash()

        return tx

    def test_output_script_size_limit(self):
        """Test spec 1: Output scriptPubKeys exceeding 34 bytes are invalid."""
        self.log.info("Testing output scriptPubKey size limits...")

        node = self.nodes[0]

        # Test 1.1: 34-byte P2WSH output (exactly at limit - should pass)
        witness_program_32 = b'\x00' * 32
        script_p2wsh = CScript([OP_0, witness_program_32])  # OP_0 (1 byte) + 32-byte push = 34 bytes
        assert_equal(len(script_p2wsh), 34)

        tx_valid = self.create_test_transaction(script_p2wsh)
        result = node.testmempoolaccept([tx_valid.serialize().hex()])[0]
        if not result['allowed']:
            self.log.info(f"  DEBUG: P2WSH rejection reason: {result}")
        assert_equal(result['allowed'], True)
        self.log.info("  ✓ 34-byte P2WSH output accepted")

        # Test 1.2: 35-byte P2PK output (exceeds limit - should fail)
        pubkey_33 = b'\x02' + b'\x00' * 32  # Compressed pubkey
        script_p2pk = CScript([pubkey_33, OP_CHECKSIG])  # 33-byte push + OP_CHECKSIG = 35 bytes
        assert_equal(len(script_p2pk), 35)

        tx_invalid = self.create_test_transaction(script_p2pk)
        result = node.testmempoolaccept([tx_invalid.serialize().hex()])[0]
        assert_equal(result['allowed'], False)
        assert 'bad-txns-vout-script-toolarge' in result['reject-reason']
        self.log.info("  ✓ 35-byte P2PK output rejected")

        # Test 1.3: 37-byte bare multisig (exceeds limit - should fail)
        script_bare_multisig = CScript([OP_1, pubkey_33, OP_1, OP_CHECKMULTISIG])
        assert len(script_bare_multisig) >= 37

        tx_invalid = self.create_test_transaction(script_bare_multisig)
        result = node.testmempoolaccept([tx_invalid.serialize().hex()])[0]
        assert_equal(result['allowed'], False)
        assert 'bad-txns-vout-script-toolarge' in result['reject-reason']
        self.log.info("  ✓ 37-byte bare multisig output rejected")

        # Test 1.4: OP_RETURN with 83 bytes (at the OP_RETURN exception limit)
        # Note: CScript adds PUSHDATA overhead for data >75 bytes
        # 80 bytes data: OP_RETURN (1) + direct push (1) + data (80) = 82 bytes total
        # 81+ bytes data: OP_RETURN (1) + OP_PUSHDATA1 (1) + len (1) + data = 84+ bytes
        data_80 = b'\x00' * 80
        script_opreturn_82 = CScript([OP_RETURN, data_80])
        self.log.info(f"  DEBUG: OP_RETURN script with 80 data bytes has length: {len(script_opreturn_82)}")

        tx_valid = self.create_test_transaction(script_opreturn_82, value=0)
        result = node.testmempoolaccept([tx_valid.serialize().hex()])[0]
        # OP_RETURN with value=0 may be rejected by standardness policy
        self.log.info(f"  ✓ OP_RETURN with {len(script_opreturn_82)} bytes: {result.get('allowed', False)}")

        # Test 1.5: OP_RETURN with 85 bytes (exceeds 83-byte exception)
        data_82 = b'\x00' * 82
        script_opreturn_85 = CScript([OP_RETURN, data_82])
        self.log.info(f"  DEBUG: OP_RETURN script with 82 data bytes has length: {len(script_opreturn_85)}")

        tx_invalid = self.create_test_transaction(script_opreturn_85, value=0)
        result = node.testmempoolaccept([tx_invalid.serialize().hex()])[0]
        assert_equal(result['allowed'], False)
        if result['allowed'] == False:
            self.log.info(f"  ✓ OP_RETURN with {len(script_opreturn_85)} bytes rejected")

    def test_pushdata_size_limit(self):
        """Test spec 2: OP_PUSHDATA* with payloads > 256 bytes are invalid."""
        self.log.info("Testing OP_PUSHDATA size limits...")

        node = self.nodes[0]

        # Standard P2WPKH hash for outputs (avoids tx-size-small policy rejection)
        dummy_pubkey_hash = hash160(b'\x00' * 33)

        # Test 2.1: Witness script with 256-byte PUSHDATA (exactly at limit - should pass)
        data_256 = b'\x00' * 256
        witness_script_256 = CScript([data_256, OP_DROP, OP_TRUE])  # Script: <256 bytes> DROP TRUE
        script_pubkey_256 = script_to_p2wsh_script(witness_script_256)

        # First create an output with this witness script
        funding_tx_256 = self.create_test_transaction(script_pubkey_256)
        txid_256 = node.sendrawtransaction(funding_tx_256.serialize().hex())
        self.generate(node, 1)
        output_value_256 = funding_tx_256.vout[0].nValue

        # Now spend it - this reveals the witness script with the 256-byte PUSHDATA
        spending_tx_256 = CTransaction()
        spending_tx_256.vin = [CTxIn(COutPoint(int(txid_256, 16), 0))]
        spending_tx_256.vout = [CTxOut(output_value_256 - 10000, CScript([OP_0, dummy_pubkey_hash]))]
        spending_tx_256.wit.vtxinwit = [CTxInWitness()]
        spending_tx_256.wit.vtxinwit[0].scriptWitness.stack = [witness_script_256]
        spending_tx_256.rehash()

        # 256 bytes is at the limit, should be accepted
        result = node.testmempoolaccept([spending_tx_256.serialize().hex()])[0]
        if not result['allowed']:
            self.log.info(f"  DEBUG: 256-byte PUSHDATA rejection: {result}")
        assert_equal(result['allowed'], True)
        self.log.info("  ✓ PUSHDATA with 256 bytes accepted in witness script")

        # Test 2.2: Witness script with 257-byte PUSHDATA (exceeds limit - should fail)
        data_257 = b'\x00' * 257
        witness_script_257 = CScript([data_257, OP_DROP, OP_TRUE])
        script_pubkey_257 = script_to_p2wsh_script(witness_script_257)

        # Create and fund the output
        funding_tx_257 = self.create_test_transaction(script_pubkey_257)
        txid_257 = node.sendrawtransaction(funding_tx_257.serialize().hex())
        self.generate(node, 1)
        output_value_257 = funding_tx_257.vout[0].nValue

        # Try to spend it - should be rejected due to 257-byte PUSHDATA
        spending_tx_257 = CTransaction()
        spending_tx_257.vin = [CTxIn(COutPoint(int(txid_257, 16), 0))]
        spending_tx_257.vout = [CTxOut(output_value_257 - 10000, CScript([OP_0, dummy_pubkey_hash]))]
        spending_tx_257.wit.vtxinwit = [CTxInWitness()]
        spending_tx_257.wit.vtxinwit[0].scriptWitness.stack = [witness_script_257]
        spending_tx_257.rehash()

        result = node.testmempoolaccept([spending_tx_257.serialize().hex()])[0]
        assert_equal(result['allowed'], False)
        assert 'non-mandatory-script-verify-flag' in result['reject-reason'] or 'Push value size limit exceeded' in result['reject-reason']
        self.log.info("  ✓ PUSHDATA with 257 bytes rejected in witness script")

        # Test 2.3: P2SH redeemScript with 300-byte PUSHDATA (tests BIP16 exception boundary)
        # Important: BIP16 allows pushing the redeemScript itself even if >256 bytes,
        # BUT any PUSHDATAs executed WITHIN that redeemScript are still limited to 256 bytes
        large_redeem_script = CScript([b'\x00' * 300, OP_DROP, OP_TRUE])  # Contains 300-byte PUSHDATA
        p2sh_script_pubkey = script_to_p2sh_script(large_redeem_script)

        # Create the P2SH output
        funding_tx_p2sh = self.create_test_transaction(p2sh_script_pubkey)
        txid_p2sh = node.sendrawtransaction(funding_tx_p2sh.serialize().hex())
        self.generate(node, 1)
        output_value_p2sh = funding_tx_p2sh.vout[0].nValue

        # Spend it by revealing the redeemScript in scriptSig
        spending_tx_p2sh = CTransaction()
        spending_tx_p2sh.vin = [CTxIn(COutPoint(int(txid_p2sh, 16), 0), CScript([large_redeem_script]))]
        spending_tx_p2sh.vout = [CTxOut(output_value_p2sh - 10000, CScript([OP_0, dummy_pubkey_hash]))]
        spending_tx_p2sh.rehash()

        # Should fail because the 300-byte PUSHDATA inside the redeemScript exceeds the limit
        result = node.testmempoolaccept([spending_tx_p2sh.serialize().hex()])[0]
        assert_equal(result['allowed'], False)
        assert 'non-mandatory-script-verify-flag' in result['reject-reason'] or 'Push value size limit exceeded' in result['reject-reason']
        self.log.info("  ✓ P2SH redeemScript with >256 byte PUSHDATA correctly rejected")
        self.log.info("    (BIP16 exception only applies to pushing the redeemScript blob, not PUSHDATAs within it)")

    def test_undefined_witness_versions(self):
        """Test spec 3: Spending undefined witness versions is invalid.

        Bitcoin currently defines witness v0 (P2WPKH/P2WSH) and v1 (Taproot).
        Versions v2-v16 are reserved for future upgrades and are currently undefined.
        After DEPLOYMENT_REDUCED_DATA, spending these undefined versions is invalid.
        """
        self.log.info("Testing undefined witness version rejection...")

        node = self.nodes[0]

        # Test witness v2 as representative (same logic applies to v3-v16)
        version_op = OP_2  # Witness version 2
        version = version_op - 0x50  # Convert OP_2 to numeric 2

        # Create output to witness v2: <version> <32-byte program>
        witness_program = b'\x00' * 32
        script_v2 = CScript([CScriptOp(version_op), witness_program])

        # Step 1: Create an output to witness v2 (this is allowed)
        funding_tx = self.create_test_transaction(script_v2)
        txid = node.sendrawtransaction(funding_tx.serialize().hex())
        self.generate(node, 1)
        self.log.info(f"  Created witness v2 output in tx {txid[:16]}...")

        # Step 2: Try to spend the witness v2 output (should be rejected)
        spending_tx = CTransaction()
        spending_tx.vin = [CTxIn(COutPoint(int(txid, 16), 0))]
        dummy_pubkey_hash = hash160(b'\x00' * 33)
        spending_tx.vout = [CTxOut(funding_tx.vout[0].nValue - 10000, CScript([OP_0, dummy_pubkey_hash]))]

        # For undefined witness versions, pre-softfork behavior was "anyone-can-spend"
        # with an empty witness stack. Post-REDUCED_DATA, this is now invalid.
        spending_tx.wit.vtxinwit = [CTxInWitness()]
        spending_tx.wit.vtxinwit[0].scriptWitness.stack = []  # Empty witness
        spending_tx.rehash()

        # Should be rejected - undefined witness versions can't be spent after activation
        result = node.testmempoolaccept([spending_tx.serialize().hex()])[0]
        assert_equal(result['allowed'], False)
        # Rejection happens during script verification
        assert any(x in result['reject-reason'] for x in ['mempool-script-verify-flag', 'witness-program', 'bad-witness', 'discouraged'])
        self.log.info(f"  ✓ Witness v{version} spending correctly rejected ({result['reject-reason']})")

        # All undefined versions (v2-v16) are validated identically
        self.log.info("  ✓ Witness versions v2-v16 are all similarly rejected")

    def test_taproot_annex_rejection(self):
        """Test spec 4: Witness stacks with a Taproot annex are invalid."""
        self.log.info("Testing Taproot annex rejection...")
        node = self.nodes[0]

        # Generate a Taproot key pair for testing
        privkey = generate_privkey()
        internal_pubkey, _ = compute_xonly_pubkey(privkey)

        # Create a simple Taproot output (key-path only, no script tree)
        taproot_info = taproot_construct(internal_pubkey)
        taproot_spk = taproot_info.scriptPubKey

        # Test 4.1: Taproot key-path spend WITHOUT annex (valid baseline)
        self.log.info("  Test 4.1: Taproot key-path spend without annex (should be valid)")

        # Create funding transaction with Taproot output
        funding_tx = self.create_test_transaction(taproot_spk)
        funding_txid = funding_tx.rehash()

        # Mine the funding transaction in a block
        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Create spending transaction (key-path, no annex)
        spending_tx = CTransaction()
        spending_tx.vin = [CTxIn(COutPoint(int(funding_txid, 16), 0), nSequence=0)]
        # Use the actual output value from funding_tx minus a small fee
        output_value = funding_tx.vout[0].nValue - 1000  # 1000 sats fee
        spending_tx.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]  # P2WPKH output

        # Sign with Schnorr signature for Taproot key-path spend
        sighash = TaprootSignatureHash(spending_tx, [funding_tx.vout[0]], SIGHASH_DEFAULT, 0)
        tweaked_privkey = tweak_add_privkey(privkey, taproot_info.tweak)
        sig = sign_schnorr(tweaked_privkey, sighash)

        # Witness for key-path: just the signature
        spending_tx.wit.vtxinwit.append(CTxInWitness())
        spending_tx.wit.vtxinwit[0].scriptWitness.stack = [sig]

        # This should be accepted (no annex)
        result = node.testmempoolaccept([spending_tx.serialize().hex()])[0]
        if not result['allowed']:
            self.log.info(f"  DEBUG: Taproot spend rejection: {result}")
        assert_equal(result['allowed'], True)
        self.log.info("  ✓ Taproot key-path spend without annex: ACCEPTED")

        # Test 4.2: Taproot key-path spend WITH annex (invalid after DEPLOYMENT_REDUCED_DATA)
        self.log.info("  Test 4.2: Taproot key-path spend with annex (should be rejected)")

        # Create another funding transaction
        funding_tx2 = self.create_test_transaction(taproot_spk)
        funding_txid2 = funding_tx2.rehash()

        # Mine the funding transaction in a block
        block_height2 = node.getblockcount() + 1
        block2 = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height2), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block2.vtx.append(funding_tx2)
        add_witness_commitment(block2)
        block2.solve()
        node.submitblock(block2.serialize().hex())

        # Create spending transaction with annex
        spending_tx2 = CTransaction()
        spending_tx2.vin = [CTxIn(COutPoint(int(funding_txid2, 16), 0), nSequence=0)]
        output_value2 = funding_tx2.vout[0].nValue - 1000
        spending_tx2.vout = [CTxOut(output_value2, CScript([OP_1, bytes(20)]))]

        # Sign the transaction (annex affects sighash)
        annex = bytes([ANNEX_TAG]) + b'\x00' * 10  # Annex must start with 0x50
        sighash2 = TaprootSignatureHash(spending_tx2, [funding_tx2.vout[0]], SIGHASH_DEFAULT, 0, annex=annex)
        sig2 = sign_schnorr(tweaked_privkey, sighash2)

        # Witness for key-path with annex: [signature, annex]
        spending_tx2.wit.vtxinwit.append(CTxInWitness())
        spending_tx2.wit.vtxinwit[0].scriptWitness.stack = [sig2, annex]

        # This should be rejected (annex present)
        result2 = node.testmempoolaccept([spending_tx2.serialize().hex()])[0]
        if result2['allowed']:
            self.log.info(f"  DEBUG: Taproot spend with annex was unexpectedly accepted: {result2}")
        assert_equal(result2['allowed'], False)
        self.log.info(f"  ✓ Taproot spend with annex: REJECTED ({result2['reject-reason']})")

    def test_taproot_control_block_size(self):
        """Test spec 5: Taproot control blocks > 257 bytes are invalid."""
        self.log.info("Testing Taproot control block size limits...")
        node = self.nodes[0]

        # Control block size = 33 + 32 * num_nodes
        # Max allowed: 7 nodes = 33 + 32*7 = 257 bytes (depth 7, 128 leaves)
        # Invalid: 8 nodes = 33 + 32*8 = 289 bytes (depth 8, 256 leaves)

        max_valid_size = TAPROOT_CONTROL_MAX_SIZE_REDUCED
        assert_equal(max_valid_size, 257)
        self.log.info(f"  Max valid control block size: {max_valid_size} bytes (7 nodes)")

        # Helper function to build a balanced binary tree of given depth
        def build_tree(depth, leaf_prefix="leaf"):
            """Build a balanced binary tree for Taproot script tree."""
            if depth == 0:
                # At leaf level, return a simple script
                return (f"{leaf_prefix}", CScript([OP_TRUE]))
            else:
                # Recursively build left and right subtrees
                left = build_tree(depth - 1, f"{leaf_prefix}_L")
                right = build_tree(depth - 1, f"{leaf_prefix}_R")
                return [left, right]

        # Generate a Taproot key pair
        privkey = generate_privkey()
        internal_pubkey, _ = compute_xonly_pubkey(privkey)

        # Test 5.1: Control block with 7 merkle nodes (valid, 257 bytes)
        self.log.info("  Test 5.1: Control block with 7 nodes / depth 7 (should be valid)")

        # Build a balanced tree of depth 7 (128 leaves)
        tree_valid = build_tree(7)
        taproot_info_valid = taproot_construct(internal_pubkey, [tree_valid])
        taproot_spk_valid = taproot_info_valid.scriptPubKey

        # Create and mine funding transaction
        funding_tx_valid = self.create_test_transaction(taproot_spk_valid)
        funding_txid_valid = funding_tx_valid.rehash()

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx_valid)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Spend using the deepest leaf (which will have the longest control block)
        # The deepest leaf should be at path L_L_L_L_L_L_L (all left)
        deepest_leaf_name = "leaf" + "_L" * 7
        leaf_info_valid = taproot_info_valid.leaves[deepest_leaf_name]
        control_block_valid = bytes([leaf_info_valid.version + taproot_info_valid.negflag]) + internal_pubkey + leaf_info_valid.merklebranch

        # Verify control block size
        assert_equal(len(control_block_valid), 257)
        self.log.info(f"    Control block size: {len(control_block_valid)} bytes ✓")

        # Create spending transaction
        spending_tx_valid = CTransaction()
        spending_tx_valid.vin = [CTxIn(COutPoint(int(funding_txid_valid, 16), 0), nSequence=0)]
        output_value = funding_tx_valid.vout[0].nValue - 1000
        spending_tx_valid.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

        spending_tx_valid.wit.vtxinwit.append(CTxInWitness())
        spending_tx_valid.wit.vtxinwit[0].scriptWitness.stack = [leaf_info_valid.script, control_block_valid]

        result_valid = node.testmempoolaccept([spending_tx_valid.serialize().hex()])[0]
        if not result_valid['allowed']:
            self.log.info(f"    DEBUG: Depth 7 rejection: {result_valid}")
        assert_equal(result_valid['allowed'], True)
        self.log.info("  ✓ Control block with 7 nodes (257 bytes): ACCEPTED")

        # Test 5.2: Control block with 8 merkle nodes (invalid, 289 bytes)
        self.log.info("  Test 5.2: Control block with 8 nodes / depth 8 (should be rejected)")

        # Build a balanced tree of depth 8 (256 leaves)
        tree_invalid = build_tree(8)
        taproot_info_invalid = taproot_construct(internal_pubkey, [tree_invalid])
        taproot_spk_invalid = taproot_info_invalid.scriptPubKey

        # Create and mine funding transaction
        funding_tx_invalid = self.create_test_transaction(taproot_spk_invalid)
        funding_txid_invalid = funding_tx_invalid.rehash()

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx_invalid)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Spend using the deepest leaf
        deepest_leaf_name_invalid = "leaf" + "_L" * 8
        leaf_info_invalid = taproot_info_invalid.leaves[deepest_leaf_name_invalid]
        control_block_invalid = bytes([leaf_info_invalid.version + taproot_info_invalid.negflag]) + internal_pubkey + leaf_info_invalid.merklebranch

        # Verify control block size
        assert_equal(len(control_block_invalid), 289)
        self.log.info(f"    Control block size: {len(control_block_invalid)} bytes (exceeds 257)")

        # Create spending transaction
        spending_tx_invalid = CTransaction()
        spending_tx_invalid.vin = [CTxIn(COutPoint(int(funding_txid_invalid, 16), 0), nSequence=0)]
        output_value = funding_tx_invalid.vout[0].nValue - 1000
        spending_tx_invalid.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

        spending_tx_invalid.wit.vtxinwit.append(CTxInWitness())
        spending_tx_invalid.wit.vtxinwit[0].scriptWitness.stack = [leaf_info_invalid.script, control_block_invalid]

        result_invalid = node.testmempoolaccept([spending_tx_invalid.serialize().hex()])[0]
        if result_invalid['allowed']:
            self.log.info(f"    DEBUG: Depth 8 was unexpectedly accepted: {result_invalid}")
        assert_equal(result_invalid['allowed'], False)
        self.log.info(f"  ✓ Control block with 8 nodes (289 bytes): REJECTED ({result_invalid['reject-reason']})")

    def test_op_success_rejection(self):
        """Test spec 6: Tapscripts including OP_SUCCESS* opcodes are invalid."""
        self.log.info("Testing OP_SUCCESS opcode rejection...")
        node = self.nodes[0]

        # Generate a Taproot key pair
        privkey = generate_privkey()
        internal_pubkey, _ = compute_xonly_pubkey(privkey)

        # Test 6.1: Tapscript without OP_SUCCESS (valid baseline)
        self.log.info("  Test 6.1: Tapscript without OP_SUCCESS (should be valid)")

        # Create a simple Tapscript: OP_TRUE (always valid)
        tapscript_valid = CScript([OP_TRUE])
        taproot_info_valid = taproot_construct(internal_pubkey, [("valid", tapscript_valid)])
        taproot_spk_valid = taproot_info_valid.scriptPubKey

        # Create and mine funding transaction
        funding_tx_valid = self.create_test_transaction(taproot_spk_valid)
        funding_txid_valid = funding_tx_valid.rehash()

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx_valid)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Create spending transaction (script-path)
        spending_tx_valid = CTransaction()
        spending_tx_valid.vin = [CTxIn(COutPoint(int(funding_txid_valid, 16), 0), nSequence=0)]
        output_value = funding_tx_valid.vout[0].nValue - 1000
        spending_tx_valid.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

        # Build witness for script-path spend
        leaf_info = taproot_info_valid.leaves["valid"]
        control_block = bytes([leaf_info.version + taproot_info_valid.negflag]) + internal_pubkey + leaf_info.merklebranch
        spending_tx_valid.wit.vtxinwit.append(CTxInWitness())
        spending_tx_valid.wit.vtxinwit[0].scriptWitness.stack = [tapscript_valid, control_block]

        result_valid = node.testmempoolaccept([spending_tx_valid.serialize().hex()])[0]
        if not result_valid['allowed']:
            self.log.info(f"  DEBUG: Valid Tapscript rejection: {result_valid}")
        assert_equal(result_valid['allowed'], True)
        self.log.info("  ✓ Tapscript without OP_SUCCESS: ACCEPTED")

        # Test 6.2: Tapscript with OP_SUCCESS (invalid)
        self.log.info("  Test 6.2: Tapscript with OP_SUCCESS (should be rejected)")

        # Create a Tapscript with OP_SUCCESS: opcodes 0x50, 0x62, etc.
        # IMPORTANT: Use CScriptOp to create the actual opcode, not PUSHDATA
        # Testing 0x50 (which is also ANNEX_TAG but different context)
        for op_success in [0x50, 0x62, 0x89]:
            tapscript_invalid = CScript([CScriptOp(op_success)])
            taproot_info_invalid = taproot_construct(internal_pubkey, [("invalid", tapscript_invalid)])
            taproot_spk_invalid = taproot_info_invalid.scriptPubKey

            # Create and mine funding transaction
            funding_tx_invalid = self.create_test_transaction(taproot_spk_invalid)
            funding_txid_invalid = funding_tx_invalid.rehash()

            block_height = node.getblockcount() + 1
            block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
            block.vtx.append(funding_tx_invalid)
            add_witness_commitment(block)
            block.solve()
            node.submitblock(block.serialize().hex())

            # Create spending transaction
            spending_tx_invalid = CTransaction()
            spending_tx_invalid.vin = [CTxIn(COutPoint(int(funding_txid_invalid, 16), 0), nSequence=0)]
            output_value = funding_tx_invalid.vout[0].nValue - 1000
            spending_tx_invalid.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

            # Build witness for script-path spend
            leaf_info_invalid = taproot_info_invalid.leaves["invalid"]
            control_block_invalid = bytes([leaf_info_invalid.version + taproot_info_invalid.negflag]) + internal_pubkey + leaf_info_invalid.merklebranch
            spending_tx_invalid.wit.vtxinwit.append(CTxInWitness())
            spending_tx_invalid.wit.vtxinwit[0].scriptWitness.stack = [tapscript_invalid, control_block_invalid]

            result_invalid = node.testmempoolaccept([spending_tx_invalid.serialize().hex()])[0]
            if result_invalid['allowed']:
                self.log.info(f"  DEBUG: OP_SUCCESS 0x{op_success:02x} was unexpectedly accepted")
            assert_equal(result_invalid['allowed'], False)
            self.log.info(f"  ✓ Tapscript with OP_SUCCESS (0x{op_success:02x}): REJECTED ({result_invalid['reject-reason']})")

    def test_op_if_notif_rejection(self):
        """Test spec 7: Tapscripts executing OP_IF or OP_NOTIF are invalid."""
        self.log.info("Testing OP_IF/OP_NOTIF rejection in Tapscript...")
        node = self.nodes[0]

        # Generate a Taproot key pair
        privkey = generate_privkey()
        internal_pubkey, _ = compute_xonly_pubkey(privkey)

        # Test 7.1: Tapscript with OP_IF (invalid in Tapscript under DEPLOYMENT_REDUCED_DATA)
        self.log.info("  Test 7.1: Tapscript with OP_IF (should be rejected)")

        # Create a Tapscript with OP_IF: OP_1 OP_IF OP_1 OP_ENDIF
        tapscript_if = CScript([OP_1, OP_IF, OP_1, OP_ENDIF])
        taproot_info_if = taproot_construct(internal_pubkey, [("with_if", tapscript_if)])
        taproot_spk_if = taproot_info_if.scriptPubKey

        # Create and mine funding transaction
        funding_tx_if = self.create_test_transaction(taproot_spk_if)
        funding_txid_if = funding_tx_if.rehash()

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx_if)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Create spending transaction
        spending_tx_if = CTransaction()
        spending_tx_if.vin = [CTxIn(COutPoint(int(funding_txid_if, 16), 0), nSequence=0)]
        output_value = funding_tx_if.vout[0].nValue - 1000
        spending_tx_if.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

        # Build witness for script-path spend
        leaf_info_if = taproot_info_if.leaves["with_if"]
        control_block_if = bytes([leaf_info_if.version + taproot_info_if.negflag]) + internal_pubkey + leaf_info_if.merklebranch
        spending_tx_if.wit.vtxinwit.append(CTxInWitness())
        spending_tx_if.wit.vtxinwit[0].scriptWitness.stack = [tapscript_if, control_block_if]

        result_if = node.testmempoolaccept([spending_tx_if.serialize().hex()])[0]
        if result_if['allowed']:
            self.log.info(f"  DEBUG: OP_IF was unexpectedly accepted: {result_if}")
        assert_equal(result_if['allowed'], False)
        self.log.info(f"  ✓ Tapscript with OP_IF: REJECTED ({result_if['reject-reason']})")

        # Test 7.2: Tapscript with OP_NOTIF (invalid in Tapscript under DEPLOYMENT_REDUCED_DATA)
        self.log.info("  Test 7.2: Tapscript with OP_NOTIF (should be rejected)")

        # Create a Tapscript with OP_NOTIF: OP_0 OP_NOTIF OP_1 OP_ENDIF
        tapscript_notif = CScript([OP_0, OP_NOTIF, OP_1, OP_ENDIF])
        taproot_info_notif = taproot_construct(internal_pubkey, [("with_notif", tapscript_notif)])
        taproot_spk_notif = taproot_info_notif.scriptPubKey

        # Create and mine funding transaction
        funding_tx_notif = self.create_test_transaction(taproot_spk_notif)
        funding_txid_notif = funding_tx_notif.rehash()

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx_notif)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Create spending transaction
        spending_tx_notif = CTransaction()
        spending_tx_notif.vin = [CTxIn(COutPoint(int(funding_txid_notif, 16), 0), nSequence=0)]
        output_value = funding_tx_notif.vout[0].nValue - 1000
        spending_tx_notif.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

        # Build witness for script-path spend
        leaf_info_notif = taproot_info_notif.leaves["with_notif"]
        control_block_notif = bytes([leaf_info_notif.version + taproot_info_notif.negflag]) + internal_pubkey + leaf_info_notif.merklebranch
        spending_tx_notif.wit.vtxinwit.append(CTxInWitness())
        spending_tx_notif.wit.vtxinwit[0].scriptWitness.stack = [tapscript_notif, control_block_notif]

        result_notif = node.testmempoolaccept([spending_tx_notif.serialize().hex()])[0]
        if result_notif['allowed']:
            self.log.info(f"  DEBUG: OP_NOTIF was unexpectedly accepted: {result_notif}")
        assert_equal(result_notif['allowed'], False)
        self.log.info(f"  ✓ Tapscript with OP_NOTIF: REJECTED ({result_notif['reject-reason']})")

    def test_mandatory_flags_cannot_be_bypassed(self):
        """Test that REDUCED_DATA consensus-mandatory flags cannot be bypassed via ignore_rejects.

        This test verifies that even though PolicyScriptChecks can be bypassed via ignore_rejects,
        the subsequent ConsensusScriptChecks enforces consensus rules and prevents invalid transactions
        from entering the mempool.
        """
        self.log.info("Testing that REDUCED_DATA rules are enforced despite ignore_rejects...")
        node = self.nodes[0]

        # Test case: Create a witness script with a 257-byte PUSHDATA (violates REDUCED_DATA)
        self.log.info("  Test: 257-byte PUSHDATA in witness script")

        # Create a P2WSH output with a witness script containing 257-byte data push
        witness_script_257 = CScript([b'\x00' * 257, OP_DROP, OP_TRUE])
        script_pubkey_257 = script_to_p2wsh_script(witness_script_257)

        # Create and fund the output
        funding_tx_257 = self.create_test_transaction(script_pubkey_257)
        txid_257 = node.sendrawtransaction(funding_tx_257.serialize().hex())
        self.generate(node, 1)
        output_value_257 = funding_tx_257.vout[0].nValue

        # Create spending transaction that reveals the 257-byte PUSHDATA
        spending_tx_257 = CTransaction()
        spending_tx_257.vin = [CTxIn(COutPoint(int(txid_257, 16), 0))]
        # Add padding to output to ensure tx meets minimum size requirements (82 bytes non-witness)
        spending_tx_257.vout = [CTxOut(output_value_257 - 1000, CScript([OP_TRUE, OP_DROP] + [OP_TRUE] * 30))]
        spending_tx_257.wit.vtxinwit.append(CTxInWitness())
        spending_tx_257.wit.vtxinwit[0].scriptWitness.stack = [witness_script_257]
        spending_tx_257.rehash()

        # Test 1: Normal testmempoolaccept should reject
        self.log.info("    Test 1a: Normal testmempoolaccept (should reject)")
        result_normal = node.testmempoolaccept([spending_tx_257.serialize().hex()])[0]
        assert_equal(result_normal['allowed'], False)
        assert 'mempool-script-verify-flag' in result_normal['reject-reason']
        self.log.info(f"    ✓ Normal testmempoolaccept correctly rejected: {result_normal['reject-reason']}")

        # Test 2: Try to bypass with ignore_rejects=["non-mandatory-script-verify-flag"]
        # Expected: Transaction is STILL REJECTED because ConsensusScriptChecks enforces consensus rules
        self.log.info("    Test 1b: testmempoolaccept with ignore_rejects")
        self.log.info("      This bypasses PolicyScriptChecks but NOT ConsensusScriptChecks")
        result_bypass = node.testmempoolaccept(
            rawtxs=[spending_tx_257.serialize().hex()],
            ignore_rejects=["mempool-script-verify-flag-failed"]
        )[0]

        # The transaction should still be rejected because ConsensusScriptChecks
        # uses GetBlockScriptFlags() which includes REDUCED_DATA consensus rules
        self.log.info(f"    Result: allowed={result_bypass['allowed']}")
        assert_equal(result_bypass['allowed'], False)
        self.log.info(f"    ✓ Transaction correctly rejected: {result_bypass['reject-reason']}")
        self.log.info("    ✓ ConsensusScriptChecks prevents bypass of REDUCED_DATA consensus rules")

    def test_p2wsh_multisig_witness_script_exemption(self):
        """Test that a large P2WSH witness script (>256 bytes) is exempted from the element size limit.

        Inspired by mainnet tx a0032427454536006263d237819df5e72fe539a38cb26264ea45a1019fb53bee,
        which is a 9-input transaction where each input spends an 11-of-15 P2WSH multisig.

        The witness script for 11-of-15 multisig is ~513 bytes, which exceeds the 256-byte
        MAX_SCRIPT_ELEMENT_SIZE_REDUCED limit. However, for P2WSH spends, the witness script
        is popped from the stack BEFORE the element size check runs in ExecuteWitnessScript,
        so it is implicitly exempted.
        """
        self.log.info("Testing 11-of-15 P2WSH multisig witness script exemption...")

        node = self.nodes[0]

        # Generate 15 key pairs
        privkeys = [generate_privkey() for _ in range(15)]
        pubkeys = []
        for priv in privkeys:
            k = ECKey()
            k.set(priv, compressed=True)
            pubkeys.append(k.get_pubkey().get_bytes())

        # Build 11-of-15 multisig witness script:
        # OP_11 <pub1> <pub2> ... <pub15> OP_15 OP_CHECKMULTISIG
        witness_script = CScript([OP_11] + pubkeys + [OP_15, OP_CHECKMULTISIG])
        self.log.info(f"  Witness script size: {len(witness_script)} bytes")
        assert len(witness_script) > MAX_SCRIPT_ELEMENT_SIZE_REDUCED, \
            f"Witness script should exceed 256 bytes, got {len(witness_script)}"

        # Create P2WSH output
        script_pubkey = script_to_p2wsh_script(witness_script)

        # Fund the P2WSH output
        funding_tx = self.create_test_transaction(script_pubkey)
        txid = node.sendrawtransaction(funding_tx.serialize().hex())
        self.generate(node, 1)

        # Create spending transaction
        spending_tx = CTransaction()
        spending_tx.vin = [CTxIn(COutPoint(int(txid, 16), 0))]
        output_value = funding_tx.vout[0].nValue - 10000
        spending_tx.vout = [CTxOut(output_value, CScript([OP_0, hash160(b'\x01' * 33)]))]

        # Sign with 11 of the 15 keys
        spending_tx.wit.vtxinwit = [CTxInWitness()]
        sighash = SegwitV0SignatureHash(
            witness_script, spending_tx, 0, SIGHASH_ALL, funding_tx.vout[0].nValue
        )

        sigs = []
        for i in range(11):
            k = ECKey()
            k.set(privkeys[i], compressed=True)
            sig = k.sign_ecdsa(sighash) + b'\x01'  # SIGHASH_ALL
            sigs.append(sig)

        # Witness stack: [OP_0_dummy, sig1, ..., sig11, witness_script]
        spending_tx.wit.vtxinwit[0].scriptWitness.stack = [b''] + sigs + [witness_script]
        spending_tx.rehash()

        # Should be ACCEPTED: witness script is popped before size check
        result = node.testmempoolaccept([spending_tx.serialize().hex()])[0]
        assert_equal(result['allowed'], True)
        self.log.info("  PASS: 11-of-15 P2WSH multisig accepted under reduced_data")

    def test_tapscript_script_exemption(self):
        """Test that a large tapleaf script (>256 bytes) is exempted from the element size limit.

        Similar to P2WSH, for tapscript spends the tapleaf script is popped from the
        witness stack BEFORE the element size check runs in ExecuteWitnessScript,
        so it is implicitly exempted.
        """
        self.log.info("Testing tapleaf script size exemption...")

        node = self.nodes[0]

        # Build a tapscript >256 bytes using repeated <data> OP_DROP, ending with OP_TRUE
        # Each data push is ≤256 bytes (valid), but the total script exceeds 256 bytes.
        large_tapscript = CScript([b'\x00' * 200, OP_DROP, b'\x00' * 200, OP_DROP, OP_TRUE])
        assert len(large_tapscript) > MAX_SCRIPT_ELEMENT_SIZE_REDUCED, \
            f"Tapscript should exceed 256 bytes, got {len(large_tapscript)}"
        self.log.info(f"  Tapleaf script size: {len(large_tapscript)} bytes")

        # Construct taproot output with this script as a leaf
        privkey = generate_privkey()
        internal_pubkey, _ = compute_xonly_pubkey(privkey)
        taproot_info = taproot_construct(internal_pubkey, [("large_script", large_tapscript)])
        taproot_spk = taproot_info.scriptPubKey

        # Fund the taproot output
        funding_tx = self.create_test_transaction(taproot_spk)
        funding_txid = funding_tx.rehash()

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(funding_tx)
        add_witness_commitment(block)
        block.solve()
        node.submitblock(block.serialize().hex())

        # Spend via script path
        leaf_info = taproot_info.leaves["large_script"]
        control_block = bytes([leaf_info.version + taproot_info.negflag]) + internal_pubkey + leaf_info.merklebranch

        spending_tx = CTransaction()
        spending_tx.vin = [CTxIn(COutPoint(int(funding_txid, 16), 0), nSequence=0)]
        output_value = funding_tx.vout[0].nValue - 1000
        spending_tx.vout = [CTxOut(output_value, CScript([OP_1, bytes(20)]))]

        # Witness stack: [<empty stack for script execution>, script, control_block]
        # The script just does <data> DROP <data> DROP TRUE, so no stack inputs needed
        spending_tx.wit.vtxinwit.append(CTxInWitness())
        spending_tx.wit.vtxinwit[0].scriptWitness.stack = [large_tapscript, control_block]

        spending_tx.rehash()

        # Should be ACCEPTED: tapleaf script is popped before size check
        result = node.testmempoolaccept([spending_tx.serialize().hex()])[0]
        assert_equal(result['allowed'], True)
        self.log.info("  PASS: >256-byte tapleaf script accepted under reduced_data")

    def test_generation_output_size_limit(self):
        """Test that generation tx outputs are also subject to output size limits."""
        self.log.info("Testing generation tx output scriptPubKey size limits...")

        node = self.nodes[0]

        def create_block_with_generation_output(script_pubkey):
            """Helper to create a block with a custom generation tx output script."""
            tip = node.getbestblockhash()
            height = node.getblockcount() + 1
            tip_header = node.getblockheader(tip)
            block_time = tip_header['time'] + 1
            coinbase = create_coinbase(height, script_pubkey=script_pubkey)
            block = create_block(int(tip, 16), coinbase, ntime=block_time)
            add_witness_commitment(block)
            block.solve()
            return block

        # Test 1: 34-byte P2WSH generation tx output (exactly at limit - should pass)
        self.log.info("  Test: 34-byte P2WSH generation tx output (at limit)")
        witness_program_32 = b'\x00' * 32
        script_p2wsh = CScript([OP_0, witness_program_32])
        assert_equal(len(script_p2wsh), 34)

        block_valid = create_block_with_generation_output(script_p2wsh)
        result = node.submitblock(block_valid.serialize().hex())
        assert_equal(result, None)
        self.log.info("  ✓ 34-byte P2WSH generation tx output accepted")

        # Test 2: 35-byte P2PK generation tx output (exceeds limit - should fail)
        self.log.info("  Test: 35-byte P2PK generation tx output (exceeds limit)")
        pubkey_33 = b'\x02' + b'\x00' * 32  # Compressed pubkey format
        script_p2pk = CScript([pubkey_33, OP_CHECKSIG])
        assert_equal(len(script_p2pk), 35)

        block_invalid = create_block_with_generation_output(script_p2pk)
        result = node.submitblock(block_invalid.serialize().hex())
        assert_equal(result, 'bad-txns-vout-script-toolarge')
        self.log.info("  ✓ 35-byte P2PK generation tx output rejected")

        # Test 3: Generation tx with OP_RETURN at 83 bytes (at OP_RETURN limit - should pass)
        self.log.info("  Test: Generation tx with 83-byte OP_RETURN extra output (at limit)")
        # 80 bytes data = OP_RETURN (1) + push opcode (1) + data (80) = 82 bytes
        # We need 83 bytes, so use 81 bytes of data with PUSHDATA1
        # OP_RETURN (1) + OP_PUSHDATA1 (1) + len (1) + data (80) = 83 bytes
        data_80 = b'\x00' * 80
        script_opreturn_83 = CScript([OP_RETURN, data_80])
        # Verify we're at exactly 83 bytes (with CScript's encoding)
        self.log.info(f"    OP_RETURN script length: {len(script_opreturn_83)} bytes")

        # Create block with valid main output and OP_RETURN extra output
        tip = node.getbestblockhash()
        height = node.getblockcount() + 1
        tip_header = node.getblockheader(tip)
        block_time = tip_header['time'] + 1
        coinbase = create_coinbase(height, extra_output_script=script_opreturn_83)
        block_opreturn_valid = create_block(int(tip, 16), coinbase, ntime=block_time)
        add_witness_commitment(block_opreturn_valid)
        block_opreturn_valid.solve()

        result = node.submitblock(block_opreturn_valid.serialize().hex())
        if result is None:
            self.log.info("  ✓ Generation tx with 83-byte OP_RETURN output accepted")
        else:
            self.log.info(f"  Note: Generation tx OP_RETURN result: {result}")

        # Test 4: Generation tx with OP_RETURN at 84 bytes (exceeds limit - should fail)
        self.log.info("  Test: Generation tx with 84-byte OP_RETURN extra output (exceeds limit)")
        # 81 bytes data = OP_RETURN (1) + OP_PUSHDATA1 (1) + len (1) + data (81) = 84 bytes
        data_81 = b'\x00' * 81
        script_opreturn_84 = CScript([OP_RETURN, data_81])
        self.log.info(f"    OP_RETURN script length: {len(script_opreturn_84)} bytes")

        tip = node.getbestblockhash()
        height = node.getblockcount() + 1
        tip_header = node.getblockheader(tip)
        block_time = tip_header['time'] + 1
        coinbase = create_coinbase(height, extra_output_script=script_opreturn_84)
        block_opreturn_invalid = create_block(int(tip, 16), coinbase, ntime=block_time)
        add_witness_commitment(block_opreturn_invalid)
        block_opreturn_invalid.solve()

        result = node.submitblock(block_opreturn_invalid.serialize().hex())
        assert_equal(result, 'bad-txns-vout-script-toolarge')
        self.log.info("  ✓ Generation tx with 84-byte OP_RETURN output rejected")

    def test_p2a_witness_rejected(self):
        """Test that P2A (PayToAnchor) spends with non-empty witness are rejected."""
        self.log.info("Testing P2A non-empty witness rejection...")
        node = self.nodes[0]

        # Create a P2A output (4 bytes, within the 34-byte limit)
        p2a_funding = self.create_test_transaction(PAY_TO_ANCHOR)
        p2a_funding.rehash()
        p2a_value = p2a_funding.vout[0].nValue

        block_height = node.getblockcount() + 1
        block = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block.vtx.append(p2a_funding)
        add_witness_commitment(block)
        block.solve()
        assert_equal(node.submitblock(block.serialize().hex()), None)
        self.log.info("  P2A output created")

        # Test 1: Spend with 100 KB of arbitrary witness data (must be rejected)
        self.log.info("  Test: P2A spend with large arbitrary witness (should be rejected)")
        arbitrary_data = b'\xab' * 100_000

        p2a_spend = CTransaction()
        p2a_spend.vin = [CTxIn(COutPoint(int(p2a_funding.rehash(), 16), 0))]
        p2a_spend.vout = [CTxOut(p2a_value - 1000, CScript([OP_0, hash160(b'\x01' * 33)]))]
        p2a_spend.wit.vtxinwit = [CTxInWitness()]
        p2a_spend.wit.vtxinwit[0].scriptWitness.stack = [arbitrary_data]
        p2a_spend.rehash()

        block_height = node.getblockcount() + 1
        block_bad = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block_bad.vtx.append(p2a_spend)
        add_witness_commitment(block_bad)
        block_bad.solve()

        result = node.submitblock(block_bad.serialize().hex())
        assert result is not None
        assert_equal(node.getblockcount(), block_height - 1)
        self.log.info(f"  ✓ P2A spend with 100 KB witness rejected ({result})")

        # Test 2: Spend with empty witness (must still be accepted)
        self.log.info("  Test: P2A spend with empty witness (should be accepted)")
        p2a_spend_empty = CTransaction()
        p2a_spend_empty.vin = [CTxIn(COutPoint(int(p2a_funding.rehash(), 16), 0))]
        p2a_spend_empty.vout = [CTxOut(p2a_value - 1000, CScript([OP_0, hash160(b'\x01' * 33)]))]
        p2a_spend_empty.wit.vtxinwit = [CTxInWitness()]
        p2a_spend_empty.wit.vtxinwit[0].scriptWitness.stack = []
        p2a_spend_empty.rehash()

        block_height = node.getblockcount() + 1
        block_good = create_block(int(node.getbestblockhash(), 16), create_coinbase(block_height), int(node.getblockheader(node.getbestblockhash())['time']) + 1)
        block_good.vtx.append(p2a_spend_empty)
        add_witness_commitment(block_good)
        block_good.solve()

        assert_equal(node.submitblock(block_good.serialize().hex()), None)
        assert_equal(node.getblockcount(), block_height)
        self.log.info("  ✓ P2A spend with empty witness accepted")

    def run_test(self):
        self.init_test()

        # Run all spec tests
        self.test_output_script_size_limit()
        self.test_generation_output_size_limit()
        self.test_pushdata_size_limit()
        self.test_undefined_witness_versions()
        self.test_taproot_annex_rejection()
        self.test_taproot_control_block_size()
        self.test_op_success_rejection()
        self.test_op_if_notif_rejection()
        self.test_mandatory_flags_cannot_be_bypassed()
        self.test_p2a_witness_rejected()
        self.test_p2wsh_multisig_witness_script_exemption()
        self.test_tapscript_script_exemption()

        self.log.info("All ReducedData tests completed")


if __name__ == '__main__':
    ReducedDataTest(__file__).main()



#!/usr/bin/env python3
# Copyright (c) 2022 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Test SENDTXRCNCL message
"""

from test_framework.messages import (
    msg_sendtxrcncl,
    msg_verack,
    msg_version,
    msg_wtxidrelay,
    NODE_BLOOM,
)
from test_framework.p2p import (
    P2PInterface,
    P2P_SERVICES,
    P2P_SUBVERSION,
    P2P_VERSION,
)
from test_framework.test_framework import BitcoinTestFramework
from test_framework.util import assert_equal

class PeerNoVerack(P2PInterface):
    def __init__(self, wtxidrelay=True):
        super().__init__(wtxidrelay=wtxidrelay)

    def on_version(self, message):
        # Avoid sending verack in response to version.
        # When calling add_p2p_connection, wait_for_verack=False must be set (see
        # comment in add_p2p_connection).
        self.send_version()
        if message.nVersion >= 70016 and self.wtxidrelay:
            self.send_message(msg_wtxidrelay())

class SendTxrcnclReceiver(P2PInterface):
    def __init__(self):
        super().__init__()
        self.sendtxrcncl_msg_received = None

    def on_sendtxrcncl(self, message):
        self.sendtxrcncl_msg_received = message


class P2PFeelerReceiver(SendTxrcnclReceiver):
    def on_version(self, message):
        # feeler connections can not send any message other than their own version
        self.send_version()


class PeerTrackMsgOrder(P2PInterface):
    def __init__(self):
        super().__init__()
        self.messages = []

    def on_message(self, message):
        super().on_message(message)
        self.messages.append(message)

def create_sendtxrcncl_msg():
    sendtxrcncl_msg = msg_sendtxrcncl()
    sendtxrcncl_msg.version = 1
    sendtxrcncl_msg.salt = 2
    return sendtxrcncl_msg

class SendTxRcnclTest(BitcoinTestFramework):
    def set_test_params(self):
        self.num_nodes = 1
        self.extra_args = [['-txreconciliation']]

    def run_test(self):
        # Check everything concerning *sending* SENDTXRCNCL
        # First, *sending* to *inbound*.
        self.log.info('SENDTXRCNCL sent to an inbound')
        peer = self.nodes[0].add_p2p_connection(SendTxrcnclReceiver(), send_version=True, wait_for_verack=True)
        assert peer.sendtxrcncl_msg_received
        assert_equal(peer.sendtxrcncl_msg_received.version, 1)
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL should be sent before VERACK')
        peer = self.nodes[0].add_p2p_connection(PeerTrackMsgOrder(), send_version=True, wait_for_verack=True)
        peer.wait_for_verack()
        verack_index = [i for i, msg in enumerate(peer.messages) if msg.msgtype == b'verack'][0]
        sendtxrcncl_index = [i for i, msg in enumerate(peer.messages) if msg.msgtype == b'sendtxrcncl'][0]
        assert sendtxrcncl_index < verack_index
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL on pre-WTXID version should not be sent')
        peer = self.nodes[0].add_p2p_connection(SendTxrcnclReceiver(), send_version=False, wait_for_verack=False)
        pre_wtxid_version_msg = msg_version()
        pre_wtxid_version_msg.nVersion = 70015
        pre_wtxid_version_msg.strSubVer = P2P_SUBVERSION
        pre_wtxid_version_msg.nServices = P2P_SERVICES
        pre_wtxid_version_msg.relay = 1
        peer.send_message(pre_wtxid_version_msg)
        peer.wait_for_verack()
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL for fRelay=false should not be sent')
        peer = self.nodes[0].add_p2p_connection(SendTxrcnclReceiver(), send_version=False, wait_for_verack=False)
        no_txrelay_version_msg = msg_version()
        no_txrelay_version_msg.nVersion = P2P_VERSION
        no_txrelay_version_msg.strSubVer = P2P_SUBVERSION
        no_txrelay_version_msg.nServices = P2P_SERVICES
        no_txrelay_version_msg.relay = 0
        peer.send_message(no_txrelay_version_msg)
        peer.wait_for_verack()
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL for fRelay=false should not be sent (with NODE_BLOOM offered)')
        self.restart_node(0, ["-peerbloomfilters", "-txreconciliation"])
        peer = self.nodes[0].add_p2p_connection(SendTxrcnclReceiver(), send_version=False, wait_for_verack=False)
        no_txrelay_version_msg = msg_version()
        no_txrelay_version_msg.nVersion = P2P_VERSION
        no_txrelay_version_msg.strSubVer = P2P_SUBVERSION
        no_txrelay_version_msg.nServices = P2P_SERVICES
        no_txrelay_version_msg.relay = 0
        peer.send_message(no_txrelay_version_msg)
        peer.wait_for_verack()
        assert peer.nServices & NODE_BLOOM != 0
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        # Now, *sending* to *outbound*.
        self.log.info('SENDTXRCNCL sent to an outbound')
        peer = self.nodes[0].add_outbound_p2p_connection(
            SendTxrcnclReceiver(), wait_for_verack=True, p2p_idx=0, connection_type="outbound-full-relay")
        assert peer.sendtxrcncl_msg_received
        assert_equal(peer.sendtxrcncl_msg_received.version, 1)
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL should not be sent if block-relay-only')
        peer = self.nodes[0].add_outbound_p2p_connection(
            SendTxrcnclReceiver(), wait_for_verack=True, p2p_idx=0, connection_type="block-relay-only")
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        self.log.info("SENDTXRCNCL should not be sent if feeler")
        peer = self.nodes[0].add_outbound_p2p_connection(P2PFeelerReceiver(), p2p_idx=0, connection_type="feeler")
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        self.log.info("SENDTXRCNCL should not be sent if addrfetch")
        peer = self.nodes[0].add_outbound_p2p_connection(
            SendTxrcnclReceiver(), wait_for_verack=True, p2p_idx=0, connection_type="addr-fetch")
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL not sent if -txreconciliation flag is not set')
        self.restart_node(0, [])
        peer = self.nodes[0].add_p2p_connection(SendTxrcnclReceiver(), send_version=True, wait_for_verack=True)
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        self.log.info('SENDTXRCNCL not sent if blocksonly is set')
        self.restart_node(0, ["-txreconciliation", "-blocksonly"])
        peer = self.nodes[0].add_p2p_connection(SendTxrcnclReceiver(), send_version=True, wait_for_verack=True)
        assert not peer.sendtxrcncl_msg_received
        self.nodes[0].disconnect_p2ps()

        # Check everything concerning *receiving* SENDTXRCNCL
        # First, receiving from *inbound*.
        self.restart_node(0, ["-txreconciliation"])
        self.log.info('valid SENDTXRCNCL received')
        peer = self.nodes[0].add_p2p_connection(PeerNoVerack(), send_version=True, wait_for_verack=False)
        with self.nodes[0].assert_debug_log(["received: sendtxrcncl"]):
            peer.send_message(create_sendtxrcncl_msg())
        self.log.info('second SENDTXRCNCL triggers a disconnect')
        with self.nodes[0].assert_debug_log(["(sendtxrcncl received from already registered peer), disconnecting peer=0"]):
            peer.send_message(create_sendtxrcncl_msg())
            peer.wait_for_disconnect()

        self.restart_node(0, [])
        self.log.info('SENDTXRCNCL if no txreconciliation supported is ignored')
        peer = self.nodes[0].add_p2p_connection(PeerNoVerack(), send_version=True, wait_for_verack=False)
        with self.nodes[0].assert_debug_log(['ignored, as our node does not have txreconciliation enabled']):
            peer.send_message(create_sendtxrcncl_msg())
        self.nodes[0].disconnect_p2ps()

        self.restart_node(0, ["-txreconciliation"])

        self.log.info('SENDTXRCNCL with version=0 triggers a disconnect')
        sendtxrcncl_low_version = create_sendtxrcncl_msg()
        sendtxrcncl_low_version.version = 0
        peer = self.nodes[0].add_p2p_connection(PeerNoVerack(), send_version=True, wait_for_verack=False)
        with self.nodes[0].assert_debug_log(["txreconciliation protocol violation"]):
            peer.send_message(sendtxrcncl_low_version)
            peer.wait_for_disconnect()

        self.log.info('SENDTXRCNCL with version=2 is valid')
        sendtxrcncl_higher_version = create_sendtxrcncl_msg()
        sendtxrcncl_higher_version.version = 2
        peer = self.nodes[0].add_p2p_connection(PeerNoVerack(), send_version=True, wait_for_verack=False)
        with self.nodes[0].assert_debug_log(['Register peer=1']):
            peer.send_message(sendtxrcncl_higher_version)
        self.nodes[0].disconnect_p2ps()

        self.log.info('unexpected SENDTXRCNCL is ignored')
        peer = self.nodes[0].add_p2p_connection(PeerNoVerack(), send_version=False, wait_for_verack=False)
        old_version_msg = msg_version()
        old_version_msg.nVersion = 70015
        old_version_msg.strSubVer = P2P_SUBVERSION
        old_version_msg.nServices = P2P_SERVICES
        old_version_msg.relay = 1
        peer.send_message(old_version_msg)
        with self.nodes[0].assert_debug_log(['Ignore unexpected txreconciliation signal']):
            peer.send_message(create_sendtxrcncl_msg())
        self.nodes[0].disconnect_p2ps()

        self.log.info('sending SENDTXRCNCL after sending VERACK triggers a disconnect')
        peer = self.nodes[0].add_p2p_connection(P2PInterface())
        with self.nodes[0].assert_debug_log(["sendtxrcncl received after verack"]):
            peer.send_message(create_sendtxrcncl_msg())
            peer.wait_for_disconnect()

        self.log.info('SENDTXRCNCL without WTXIDRELAY is ignored (recon state is erased after VERACK)')
        peer = self.nodes[0].add_p2p_connection(PeerNoVerack(wtxidrelay=False), send_version=True, wait_for_verack=False)
        with self.nodes[0].assert_debug_log(['Forget txreconciliation state of peer']):
            peer.send_message(create_sendtxrcncl_msg())
            peer.send_message(msg_verack())
        self.nodes[0].disconnect_p2ps()

        # Now, *receiving* from *outbound*.
        self.log.info('SENDTXRCNCL if block-relay-only triggers a disconnect')
        peer = self.nodes[0].add_outbound_p2p_connection(
            PeerNoVerack(), wait_for_verack=False, p2p_idx=0, connection_type="block-relay-only")
        with self.nodes[0].assert_debug_log(["we indicated no tx relay, disconnecting peer=5"]):
            peer.send_message(create_sendtxrcncl_msg())
            peer.wait_for_disconnect()


if __name__ == '__main__':
    SendTxRcnclTest(__file__).main()



#!/usr/bin/env python3
# Copyright (c) 2022-present The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.

"""  Tests the coin_selection:* tracepoint API interface.
     See https://github.com/bitcoin/bitcoin/blob/master/doc/tracing.md#context-coin_selection
"""

# Test will be skipped if we don't have bcc installed
try:
    from bcc import BPF, USDT # type: ignore[import]
except ImportError:
    pass
from test_framework.test_framework import BitcoinTestFramework
from test_framework.util import (
    assert_equal,
    assert_greater_than,
    assert_raises_rpc_error,
    bpf_cflags,
)

coinselection_tracepoints_program = """
#include <uapi/linux/ptrace.h>

#define WALLET_NAME_LENGTH 16
#define ALGO_NAME_LENGTH 16

struct event_data
{
    u8 type;
    char wallet_name[WALLET_NAME_LENGTH];

    // selected coins event
    char algo[ALGO_NAME_LENGTH];
    s64 target;
    s64 waste;
    s64 selected_value;

    // create tx event
    bool success;
    s64 fee;
    s32 change_pos;

    // aps create tx event
    bool use_aps;
};

BPF_QUEUE(coin_selection_events, struct event_data, 1024);

int trace_selected_coins(struct pt_regs *ctx) {
    struct event_data data;
    void *pwallet_name = NULL, *palgo = NULL;
    __builtin_memset(&data, 0, sizeof(data));
    data.type = 1;
    bpf_usdt_readarg(1, ctx, &pwallet_name);
    bpf_probe_read_user_str(&data.wallet_name, WALLET_NAME_LENGTH, pwallet_name);
    bpf_usdt_readarg(2, ctx, &palgo);
    bpf_probe_read_user_str(&data.algo, ALGO_NAME_LENGTH, palgo);
    bpf_usdt_readarg(3, ctx, &data.target);
    bpf_usdt_readarg(4, ctx, &data.waste);
    bpf_usdt_readarg(5, ctx, &data.selected_value);
    coin_selection_events.push(&data, 0);
    return 0;
}

int trace_normal_create_tx(struct pt_regs *ctx) {
    struct event_data data;
    void *pwallet_name = NULL;
    __builtin_memset(&data, 0, sizeof(data));
    data.type = 2;
    bpf_usdt_readarg(1, ctx, &pwallet_name);
    bpf_probe_read_user_str(&data.wallet_name, WALLET_NAME_LENGTH, pwallet_name);
    bpf_usdt_readarg(2, ctx, &data.success);
    bpf_usdt_readarg(3, ctx, &data.fee);
    bpf_usdt_readarg(4, ctx, &data.change_pos);
    coin_selection_events.push(&data, 0);
    return 0;
}

int trace_attempt_aps(struct pt_regs *ctx) {
    struct event_data data;
    void *pwallet_name = NULL;
    __builtin_memset(&data, 0, sizeof(data));
    data.type = 3;
    bpf_usdt_readarg(1, ctx, &pwallet_name);
    bpf_probe_read_user_str(&data.wallet_name, WALLET_NAME_LENGTH, pwallet_name);
    coin_selection_events.push(&data, 0);
    return 0;
}

int trace_aps_create_tx(struct pt_regs *ctx) {
    struct event_data data;
    void *pwallet_name = NULL;
    __builtin_memset(&data, 0, sizeof(data));
    data.type = 4;
    bpf_usdt_readarg(1, ctx, &pwallet_name);
    bpf_probe_read_user_str(&data.wallet_name, WALLET_NAME_LENGTH, pwallet_name);
    bpf_usdt_readarg(2, ctx, &data.use_aps);
    bpf_usdt_readarg(3, ctx, &data.success);
    bpf_usdt_readarg(4, ctx, &data.fee);
    bpf_usdt_readarg(5, ctx, &data.change_pos);
    coin_selection_events.push(&data, 0);
    return 0;
}
"""


class CoinSelectionTracepointTest(BitcoinTestFramework):
    def add_options(self, parser):
        self.add_wallet_options(parser)

    def set_test_params(self):
        self.num_nodes = 1
        self.setup_clean_chain = True

    def skip_test_if_missing_module(self):
        self.skip_if_platform_not_linux()
        self.skip_if_no_bitcoind_tracepoints()
        self.skip_if_no_python_bcc()
        self.skip_if_no_bpf_permissions()
        self.skip_if_no_wallet()

    def get_tracepoints(self, expected_types):
        events = []
        try:
            for i in range(0, len(expected_types) + 1):
                event = self.bpf["coin_selection_events"].pop()
                assert_equal(event.wallet_name.decode(), self.default_wallet_name)
                assert_equal(event.type, expected_types[i])
                events.append(event)
            else:
                # If the loop exits successfully instead of throwing a KeyError, then we have had
                # more events than expected. There should be no more than len(expected_types) events.
                assert False
        except KeyError:
            assert_equal(len(events), len(expected_types))
            return events


    def determine_selection_from_usdt(self, events):
        success = None
        use_aps = None
        algo = None
        waste = None
        change_pos = None

        is_aps = False
        sc_events = []
        for event in events:
            if event.type == 1:
                if not is_aps:
                    algo = event.algo.decode()
                    waste = event.waste
                sc_events.append(event)
            elif event.type == 2:
                success = event.success
                if not is_aps:
                    change_pos = event.change_pos
            elif event.type == 3:
                is_aps = True
            elif event.type == 4:
                assert is_aps
                if event.use_aps:
                    use_aps = True
                    assert_equal(len(sc_events), 2)
                    algo = sc_events[1].algo.decode()
                    waste = sc_events[1].waste
                    change_pos = event.change_pos
        return success, use_aps, algo, waste, change_pos

    def run_test(self):
        self.log.info("hook into the coin_selection tracepoints")
        ctx = USDT(pid=self.nodes[0].process.pid)
        ctx.enable_probe(probe="coin_selection:selected_coins", fn_name="trace_selected_coins")
        ctx.enable_probe(probe="coin_selection:normal_create_tx_internal", fn_name="trace_normal_create_tx")
        ctx.enable_probe(probe="coin_selection:attempting_aps_create_tx", fn_name="trace_attempt_aps")
        ctx.enable_probe(probe="coin_selection:aps_create_tx_internal", fn_name="trace_aps_create_tx")
        self.bpf = BPF(text=coinselection_tracepoints_program, usdt_contexts=[ctx], debug=0, cflags=bpf_cflags())

        self.log.info("Prepare wallets")
        self.generate(self.nodes[0], 101)
        wallet = self.nodes[0].get_wallet_rpc(self.default_wallet_name)

        self.log.info("Sending a transaction should result in all tracepoints")
        # We should have 5 tracepoints in the order:
        # 1. selected_coins (type 1)
        # 2. normal_create_tx_internal (type 2)
        # 3. attempting_aps_create_tx (type 3)
        # 4. selected_coins (type 1)
        # 5. aps_create_tx_internal (type 4)
        wallet.sendtoaddress(wallet.getnewaddress(), 10)
        events = self.get_tracepoints([1, 2, 3, 1, 4])
        success, use_aps, _algo, _waste, change_pos = self.determine_selection_from_usdt(events)
        assert_equal(success, True)
        assert_greater_than(change_pos, -1)

        self.log.info("Failing to fund results in 1 tracepoint")
        # We should have 1 tracepoints in the order
        # 1. normal_create_tx_internal (type 2)
        assert_raises_rpc_error(-6, "Insufficient funds", wallet.sendtoaddress, wallet.getnewaddress(), 102 * 50)
        events = self.get_tracepoints([2])
        success, use_aps, _algo, _waste, change_pos = self.determine_selection_from_usdt(events)
        assert_equal(success, False)

        self.log.info("Explicitly enabling APS results in 2 tracepoints")
        # We should have 2 tracepoints in the order
        # 1. selected_coins (type 1)
        # 2. normal_create_tx_internal (type 2)
        wallet.setwalletflag("avoid_reuse")
        wallet.sendtoaddress(address=wallet.getnewaddress(), amount=10, avoid_reuse=True)
        events = self.get_tracepoints([1, 2])
        success, use_aps, _algo, _waste, change_pos = self.determine_selection_from_usdt(events)
        assert_equal(success, True)
        assert_equal(use_aps, None)

        self.log.info("Change position is -1 if no change is created with APS when APS was initially not used")
        # We should have 2 tracepoints in the order:
        # 1. selected_coins (type 1)
        # 2. normal_create_tx_internal (type 2)
        # 3. attempting_aps_create_tx (type 3)
        # 4. selected_coins (type 1)
        # 5. aps_create_tx_internal (type 4)
        wallet.sendtoaddress(address=wallet.getnewaddress(), amount=wallet.getbalance(), subtractfeefromamount=True, avoid_reuse=False)
        events = self.get_tracepoints([1, 2, 3, 1, 4])
        success, use_aps, _algo, _waste, change_pos = self.determine_selection_from_usdt(events)
        assert_equal(success, True)
        assert_equal(change_pos, -1)

        self.log.info("Change position is -1 if no change is created normally and APS is not used")
        # We should have 2 tracepoints in the order:
        # 1. selected_coins (type 1)
        # 2. normal_create_tx_internal (type 2)
        wallet.sendtoaddress(address=wallet.getnewaddress(), amount=wallet.getbalance(), subtractfeefromamount=True)
        events = self.get_tracepoints([1, 2])
        success, use_aps, _algo, _waste, change_pos = self.determine_selection_from_usdt(events)
        assert_equal(success, True)
        assert_equal(change_pos, -1)

        self.bpf.cleanup()


if __name__ == '__main__':
    CoinSelectionTracepointTest(__file__).main()



#!/usr/bin/env python3
# Copyright (c) 2018-2022 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.

import os
import sys
import argparse
import json

def perform_pre_checks():
    mock_result_path = os.path.join(os.getcwd(), "mock_result")
    if os.path.isfile(mock_result_path):
        with open(mock_result_path, "r", encoding="utf8") as f:
            mock_result = f.read()
        if mock_result[0]:
            sys.stdout.write(mock_result[2:])
            sys.exit(int(mock_result[0]))

def enumerate(args):
    sys.stdout.write(json.dumps([{"fingerprint": "00000001", "type": "trezor", "model": "trezor_t"}]))

def getdescriptors(args):
    xpub = "tpubD6NzVbkrYhZ4WaWSyoBvQwbpLkojyoTZPRsgXELWz3Popb3qkjcJyJUGLnL4qHHoQvao8ESaAstxYSnhyswJ76uZPStJRJCTKvosUCJZL5B"

    sys.stdout.write(json.dumps({
        "receive": [
            "pkh([00000001/44h/1h/" + args.account + "']" + xpub + "/0/*)#aqllu46s",
            "sh(wpkh([00000001/49h/1h/" + args.account + "']" + xpub + "/0/*))#5dh56mgg",
            "wpkh([00000001/84h/1h/" + args.account + "']" + xpub + "/0/*)#h62dxaej",
            "tr([00000001/86h/1h/" + args.account + "']" + xpub + "/0/*)#pcd5w87f"
        ],
        "internal": [
            "pkh([00000001/44h/1h/" + args.account + "']" + xpub + "/1/*)#v567pq2g",
            "sh(wpkh([00000001/49h/1h/" + args.account + "']" + xpub + "/1/*))#pvezzyah",
            "wpkh([00000001/84h/1h/" + args.account + "']" + xpub + "/1/*)#xw0vmgf2",
            "tr([00000001/86h/1h/" + args.account + "']" + xpub + "/1/*)#svg4njw3"

        ]
    }))


def displayaddress(args):
    if args.fingerprint != "00000001":
        return sys.stdout.write(json.dumps({"error": "Unexpected fingerprint", "fingerprint": args.fingerprint}))

    expected_desc = {
        "wpkh([00000001/84h/1h/0h/0/0]02c97dc3f4420402e01a113984311bf4a1b8de376cac0bdcfaf1b3ac81f13433c7)#3te6hhy7": "bcrt1qm90ugl4d48jv8n6e5t9ln6t9zlpm5th68x4f8g",
        "sh(wpkh([00000001/49h/1h/0h/0/0]02c97dc3f4420402e01a113984311bf4a1b8de376cac0bdcfaf1b3ac81f13433c7))#kz9y5w82": "2N2gQKzjUe47gM8p1JZxaAkTcoHPXV6YyVp",
        "pkh([00000001/44h/1h/0h/0/0]02c97dc3f4420402e01a113984311bf4a1b8de376cac0bdcfaf1b3ac81f13433c7)#q3pqd8wh": "n1LKejAadN6hg2FrBXoU1KrwX4uK16mco9",
        "tr([00000001/86h/1h/0h/0/0]c97dc3f4420402e01a113984311bf4a1b8de376cac0bdcfaf1b3ac81f13433c7)#puqqa90m": "tb1phw4cgpt6cd30kz9k4wkpwm872cdvhss29jga2xpmftelhqll62mscq0k4g",
        "wpkh([00000001/84h/1h/0h/0/1]03a20a46308be0b8ded6dff0a22b10b4245c587ccf23f3b4a303885be3a524f172)#aqpjv5xr": "wrong_address",
    }
    if args.desc not in expected_desc:
        return sys.stdout.write(json.dumps({"error": "Unexpected descriptor", "desc": args.desc}))

    return sys.stdout.write(json.dumps({"address": expected_desc[args.desc]}))

def signtx(args):
    if args.fingerprint != "00000001":
        return sys.stdout.write(json.dumps({"error": "Unexpected fingerprint", "fingerprint": args.fingerprint}))

    with open(os.path.join(os.getcwd(), "mock_psbt"), "r", encoding="utf8") as f:
        mock_psbt = f.read()

    if args.fingerprint == "00000001" :
        sys.stdout.write(json.dumps({
            "psbt": mock_psbt,
            "complete": True
        }))
    else:
        sys.stdout.write(json.dumps({"psbt": args.psbt}))

parser = argparse.ArgumentParser(prog='./signer.py', description='External signer mock')
parser.add_argument('--fingerprint')
parser.add_argument('--chain', default='main')
parser.add_argument('--stdin', action='store_true')

subparsers = parser.add_subparsers(description='Commands', dest='command')
subparsers.required = True

parser_enumerate = subparsers.add_parser('enumerate', help='list available signers')
parser_enumerate.set_defaults(func=enumerate)

parser_getdescriptors = subparsers.add_parser('getdescriptors')
parser_getdescriptors.set_defaults(func=getdescriptors)
parser_getdescriptors.add_argument('--account', metavar='account')

parser_displayaddress = subparsers.add_parser('displayaddress', help='display address on signer')
parser_displayaddress.add_argument('--desc', metavar='desc')
parser_displayaddress.set_defaults(func=displayaddress)

parser_signtx = subparsers.add_parser('signtx')
parser_signtx.add_argument('psbt', metavar='psbt')

parser_signtx.set_defaults(func=signtx)

if not sys.stdin.isatty():
    buffer = sys.stdin.read()
    if buffer and buffer.rstrip() != "":
        sys.argv.extend(buffer.rstrip().split(" "))

args = parser.parse_args()

perform_pre_checks()

args.func(args)



#!/usr/bin/env python3
# Copyright (c) 2014-2022 The Bitcoin Core developers
# Distributed under the MIT software license, see the accompanying
# file COPYING or http://www.opensource.org/licenses/mit-license.php.
"""Test mempool re-org scenarios.

Test re-org scenarios with a mempool that contains transactions
that spend (directly or indirectly) coinbase transactions.
"""

import time

from test_framework.messages import (
    CInv,
    MSG_WTX,
    msg_getdata,
)
from test_framework.p2p import (
    P2PTxInvStore,
    p2p_lock,
)
from test_framework.test_framework import BitcoinTestFramework
from test_framework.util import assert_equal, assert_raises_rpc_error
from test_framework.wallet import MiniWallet

class MempoolCoinbaseTest(BitcoinTestFramework):
    def set_test_params(self):
        self.num_nodes = 2
        self.extra_args = [
            [
                '-whitelist=noban@127.0.0.1',  # immediate tx relay
            ],
            []
        ]

    def test_reorg_relay(self):
        self.log.info("Test that transactions from disconnected blocks are available for relay immediately")
        # Prevent time from moving forward
        self.nodes[1].setmocktime(int(time.time()))
        self.connect_nodes(0, 1)
        self.generate(self.wallet, 3)

        # Disconnect node0 and node1 to create different chains.
        self.disconnect_nodes(0, 1)
        # Connect a peer to node1, which doesn't have immediate tx relay
        peer1 = self.nodes[1].add_p2p_connection(P2PTxInvStore())

        # Create a transaction that is included in a block.
        tx_disconnected = self.wallet.send_self_transfer(from_node=self.nodes[1])
        self.generate(self.nodes[1], 1, sync_fun=self.no_op)

        # Create a transaction and submit it to node1's mempool.
        tx_before_reorg = self.wallet.send_self_transfer(from_node=self.nodes[1])

        # Create a child of that transaction and submit it to node1's mempool.
        tx_child = self.wallet.send_self_transfer(utxo_to_spend=tx_disconnected["new_utxo"], from_node=self.nodes[1])
        assert_equal(self.nodes[1].getmempoolentry(tx_child["txid"])["ancestorcount"], 1)
        assert_equal(len(peer1.get_invs()), 0)

        # node0 has a longer chain in which tx_disconnected was not confirmed.
        self.generate(self.nodes[0], 3, sync_fun=self.no_op)

        # Reconnect the nodes and sync chains. node0's chain should win.
        self.connect_nodes(0, 1)
        self.sync_blocks()

        # Child now has an ancestor from the disconnected block
        assert_equal(self.nodes[1].getmempoolentry(tx_child["txid"])["ancestorcount"], 2)
        assert_equal(self.nodes[1].getmempoolentry(tx_before_reorg["txid"])["ancestorcount"], 1)

        # peer1 should not have received an inv for any of the transactions during this time, as no
        # mocktime has elapsed for those transactions to be announced. Likewise, it cannot
        # request very recent, unanounced transactions.
        assert_equal(len(peer1.get_invs()), 0)
        # It's too early to request these two transactions
        requests_too_recent = msg_getdata([CInv(t=MSG_WTX, h=int(tx["tx"].getwtxid(), 16)) for tx in [tx_before_reorg, tx_child]])
        peer1.send_and_ping(requests_too_recent)
        for _ in range(len(requests_too_recent.inv)):
            peer1.sync_with_ping()
        with p2p_lock:
            assert "tx" not in peer1.last_message
            assert "notfound" in peer1.last_message

        # Request the tx from the disconnected block
        request_disconnected_tx = msg_getdata([CInv(t=MSG_WTX, h=int(tx_disconnected["tx"].getwtxid(), 16))])
        peer1.send_and_ping(request_disconnected_tx)

        # The tx from the disconnected block was never announced, and it entered the mempool later
        # than the transactions that are too recent.
        assert_equal(len(peer1.get_invs()), 0)
        with p2p_lock:
            # However, the node will answer requests for the tx from the recently-disconnected block.
            assert_equal(peer1.last_message["tx"].tx.getwtxid(),tx_disconnected["tx"].getwtxid())

        self.nodes[1].setmocktime(int(time.time()) + 300)
        peer1.sync_with_ping()
        # the transactions are now announced
        assert_equal(len(peer1.get_invs()), 3)
        for _ in range(3):
            # make sure all tx requests have been responded to
            peer1.sync_with_ping()
        last_tx_received = peer1.last_message["tx"]

        tx_after_reorg = self.wallet.send_self_transfer(from_node=self.nodes[1])
        request_after_reorg = msg_getdata([CInv(t=MSG_WTX, h=int(tx_after_reorg["tx"].getwtxid(), 16))])
        assert tx_after_reorg["txid"] in self.nodes[1].getrawmempool()
        peer1.send_and_ping(request_after_reorg)
        with p2p_lock:
            assert_equal(peer1.last_message["tx"], last_tx_received)

    def run_test(self):
        self.wallet = MiniWallet(self.nodes[0])
        wallet = self.wallet

        # Start with a 200 block chain
        assert_equal(self.nodes[0].getblockcount(), 200)

        self.log.info("Add 4 coinbase utxos to the miniwallet")
        # Block 76 contains the first spendable coinbase txs.
        first_block = 76

        # Three scenarios for re-orging coinbase spends in the memory pool:
        # 1. Direct coinbase spend  :  spend_1
        # 2. Indirect (coinbase spend in chain, child in mempool) : spend_2 and spend_2_1
        # 3. Indirect (coinbase and child both in chain) : spend_3 and spend_3_1
        # Use invalidateblock to make all of the above coinbase spends invalid (immature coinbase),
        # and make sure the mempool code behaves correctly.
        b = [self.nodes[0].getblockhash(n) for n in range(first_block, first_block+4)]
        coinbase_txids = [self.nodes[0].getblock(h)['tx'][0] for h in b]
        utxo_1 = wallet.get_utxo(txid=coinbase_txids[1])
        utxo_2 = wallet.get_utxo(txid=coinbase_txids[2])
        utxo_3 = wallet.get_utxo(txid=coinbase_txids[3])
        self.log.info("Create three transactions spending from coinbase utxos: spend_1, spend_2, spend_3")
        spend_1 = wallet.create_self_transfer(utxo_to_spend=utxo_1)
        spend_2 = wallet.create_self_transfer(utxo_to_spend=utxo_2)
        spend_3 = wallet.create_self_transfer(utxo_to_spend=utxo_3)

        self.log.info("Create another transaction which is time-locked to two blocks in the future")
        utxo = wallet.get_utxo(txid=coinbase_txids[0])
        timelock_tx = wallet.create_self_transfer(
            utxo_to_spend=utxo,
            locktime=self.nodes[0].getblockcount() + 2,
        )['hex']

        self.log.info("Check that the time-locked transaction is too immature to spend")
        assert_raises_rpc_error(-26, "non-final", self.nodes[0].sendrawtransaction, timelock_tx)

        self.log.info("Broadcast and mine spend_2 and spend_3")
        wallet.sendrawtransaction(from_node=self.nodes[0], tx_hex=spend_2['hex'])
        wallet.sendrawtransaction(from_node=self.nodes[0], tx_hex=spend_3['hex'])
        self.log.info("Generate a block")
        self.generate(self.nodes[0], 1)
        self.log.info("Check that time-locked transaction is still too immature to spend")
        assert_raises_rpc_error(-26, 'non-final', self.nodes[0].sendrawtransaction, timelock_tx)

        self.log.info("Create spend_2_1 and spend_3_1")
        spend_2_1 = wallet.create_self_transfer(utxo_to_spend=spend_2["new_utxo"])
        spend_3_1 = wallet.create_self_transfer(utxo_to_spend=spend_3["new_utxo"])

        self.log.info("Broadcast and mine spend_3_1")
        spend_3_1_id = self.nodes[0].sendrawtransaction(spend_3_1['hex'])
        self.log.info("Generate a block")
        last_block = self.generate(self.nodes[0], 1)
        # generate() implicitly syncs blocks, so that peer 1 gets the block before timelock_tx
        # Otherwise, peer 1 would put the timelock_tx in m_lazy_recent_rejects

        self.log.info("The time-locked transaction can now be spent")
        timelock_tx_id = self.nodes[0].sendrawtransaction(timelock_tx)

        self.log.info("Add spend_1 and spend_2_1 to the mempool")
        spend_1_id = self.nodes[0].sendrawtransaction(spend_1['hex'])
        spend_2_1_id = self.nodes[0].sendrawtransaction(spend_2_1['hex'])

        assert_equal(set(self.nodes[0].getrawmempool()), {spend_1_id, spend_2_1_id, timelock_tx_id})
        self.sync_all()

        self.log.info("invalidate the last block")
        for node in self.nodes:
            node.invalidateblock(last_block[0])
        self.log.info("The time-locked transaction is now too immature and has been removed from the mempool")
        self.log.info("spend_3_1 has been re-orged out of the chain and is back in the mempool")
        assert_equal(set(self.nodes[0].getrawmempool()), {spend_1_id, spend_2_1_id, spend_3_1_id})

        self.log.info("Use invalidateblock to re-org back and make all those coinbase spends immature/invalid")
        b = self.nodes[0].getblockhash(first_block + 100)
        for node in self.nodes:
            node.invalidateblock(b)

        self.log.info("Check that the mempool is empty")
        assert_equal(set(self.nodes[0].getrawmempool()), set())
        self.sync_all()

        self.test_reorg_relay()


if __name__ == '__main__':
    MempoolCoinbaseTest(__file__).main()
